module mdtapers_mod
  ! taper types: 
  ! 1) 'zero'            --  needed: 'nothing' 
  ! 2) 'rect'            --  needed: t1,t2 
  ! 3) 'linear_up'       --  needed: t1,t2
  ! 4) 'linear_down'     --  needed: t1,t2
  ! 5) 'pyramid'         --  needed: t1,t2
  ! 6) 'pyramid_band'    --  needed: t1,t2,t3,t4
  ! 7) 'hann'            --  needed: t1,t2
  ! 8) 'hann_half'       --  needed: t1,t2
  ! 9) 'hann_band'       --  needed: t2,t2,t3,t4
  ! 10) 'cosine'         --  needed: t1,t2
  ! 11) 'gaussian'       --  needed: t1,t2,sigma,spread
  ! 12) 'gaussian_half'  --  needed: t1,t2,sigma,spread
  ! 13) 'gaussian_block' --  needed: t1,t2,t3,t4,sigma,spread

  ! dim parameter!
  ! dim = 3; Genera rotational 3d taper
  ! dim = 2; 2d taper, rectangular in 3rd axis
  ! dim = 1; 2d taper, rectangular in 2nd axis


  ! Contains a variety of callable tapers
  ! initialize by n1, nt, t1, t2, inv
  ! n1: length of taper
  ! nt: number of tapers ( for slepian only at this point )
  ! t1: index of tie point one    ( effective shape length : t2-(t1-1) )
  ! t2: index of tie point two    ( taper(t1:t2) = taper )
  ! inv: taper = 1.-taper         ( 'invert' the taper shape )
  ! This module basically, initializes the taper_mod module,
  ! Multiple calls can construct a more complicated taper.
  !
  ! Very usefull: http://en.wikipedia.org/wiki/Window_function
  ! Last accessed: 2/15/2010

  !The following tapers have non zero boundary points:

  ! Rect-likes
  ! Gaussian-likes
  ! 

  ! The following tapers have periodic boundary points (at t1 : 0 at t2 ~: 0):
  ! cosine-like
  ! half_hann-like
  ! Linear-likes (but linear_down at t1 ~: 0 at t2 : 0)
  ! hann-like & Pyramid-likes (at t1 : 0 at t2 : 0):
  ! gaussian (at t1 ~: 0 at t2 ~: 0)


  use addnull_mod
  use   tape_mod
  use mdtape_mod
  
  implicit none
  
  ! For internal itterated use to fill taper1,2,3
  integer, private              :: n(3),c(3)
  real,    private              :: o(3),d(3)
  real, private,allocatable     :: mdtaper(:,:,:),mdtaper_com(:,:,:)

  real, parameter, private      :: pi=3.14159265


  contains
  
  ! (Re)Initialize this module
  subroutine mdtapers_init(n_in,c_in,o_in,d_in)
    integer, intent(in),optional  :: n_in(3)
    integer, intent(in),optional  :: c_in(3)
    real   , intent(in),optional  :: o_in(3),d_in(3)

    if (present( n_in)) then; n = n_in;  else; n = (/ 1 , 1 , 1  /);  end if
    if (present( c_in)) then; c = c_in;  else; c = (/ 1 , 1 , 1  /);  end if
    if (present( o_in)) then; o = o_in;  else; o = (/ 0., 0., 0. /);  end if
    if (present( d_in)) then; d = d_in;  else; d = (/ 1., 1., 1. /);  end if
    
    if (.not. allocated(mdtaper_com)) then; allocate( mdtaper_com(n(1),n(2),n(3)) ); call addnull(.false.,mdtaper_com); end if
    
  end subroutine

  subroutine mdtapers_new_init()
    if (allocated( mdtaper     )) then; deallocate( mdtaper ); end if;
    allocate( mdtaper(n(1),n(2),n(3))); call addnull(.false.,mdtaper);
  end subroutine


  ! Close this module
  subroutine mdtapers_close()
    if (allocated( mdtaper     )) then; deallocate( mdtaper    ); end if;
    if (allocated( mdtaper_com )) then; deallocate( mdtaper_com); end if;
  end subroutine
  

 ! ! To just getting a 1d taper, for external use
 ! subroutine tapers_gettaper(taper_out,inv_in,dir_in)
 !   real,intent(out)            :: taper_out(:)
 !   logical,intent(in),optional :: inv_in
 !   integer,intent(in),optional :: dir_in
 !   logical                     :: inv
 !   integer                     :: dir

 !   if (present(inv_in)) then; inv  = inv_in;  else; inv=.false.; end if
 !   if (present(dir_in)) then; dir  = dir_in;  else; dir=1      ; end if
    
 !   if (dir==1) then; taper_out = taper1; end if
 !   if (dir==2) then; taper_out = taper2; end if
 !   if (dir==3) then; taper_out = taper3; end if
 !   call tapers_invert(inv,taper_out)
    
 ! end subroutine




  ! To just getting a 1d taper, for external use
  subroutine mdtapers_gettaper(mdtaper_out,inv_in)
    real,intent(out)            :: mdtaper_out(:,:,:)
    logical,intent(in),optional :: inv_in
    logical                     :: inv

    if (present(inv_in)) then; inv  = inv_in;  else; inv=.false.; end if
    call mdtapers_invert(inv,mdtaper)
    mdtaper_out = mdtaper
    call mdtapers_invert(inv,mdtaper)

  end subroutine
  
  
  
  ! Subroutine to pass constructed taper into taper1,2,3
  subroutine  mdtapers_insert(add,t1_in,t2_in,inv_in)
    logical, intent(in)            :: add
    integer, intent(in),optional   :: t1_in(3),t2_in(3)
    logical, intent(in),optional   :: inv_in
    integer                        :: t1(3),t2(3),ii
    logical                        :: inv

    if (present(t1_in )) then; t1  = t1_in;  else; t1 = (/ 1  , 1  , 1   /);       end if
    if (present(t2_in )) then; t2  = t2_in;  else; t2 = (/n(1),n(2),n(3) /);       end if
    if (present(inv_in)) then; inv = inv_in; else; inv=.false.; end if

    call mdtapers_invert(inv,mdtaper_com)
    call addnull(add,mdtaper(t1(1):t2(1),t1(2):t2(2),t1(3):t2(3)))
    mdtaper(t1(1):t2(1),t1(2):t2(2),t1(3):t2(3)) = mdtaper(t1(1):t2(1),t1(2):t2(2),t1(3):t2(3)) + mdtaper_com(t1(1):t2(1),t1(2):t2(2),t1(3):t2(3))
    call mdtapers_invert(inv,mdtaper_com)
    
  end subroutine
  
  
  
  
  subroutine mdtapers_insert_mdtape(add_in,inv_in,t1_in,t2_in)
    logical,optional,intent(in)    :: inv_in,add_in
    integer,optional,intent(in)    :: t1_in(3),t2_in(3)
    integer                        :: t1(3),t2(3),dir
    logical                        :: inv,add
    
    integer                        :: ii,jj,kk
    
    if (present(t1_in )) then; t1  = t1_in;  else; t1 = (/ 1  , 1  , 1   /);       end if
    if (present(t2_in )) then; t2  = t2_in;  else; t2 = (/n(1),n(2),n(3) /);       end if
    if (present(inv_in)) then; inv = inv_in; else; inv=.false.; end if
    if (present(add_in)) then; add = add_in; else; add=.false.; end if
    
    call mdtapers_invert (inv,mdtaper)
    call mdtape_insert   (add,mdtaper,t1,t2)
    call mdtapers_invert (inv,mdtaper)
    
  end subroutine
  
  
  subroutine mdtapers_interface(ntapers, p1_in,p2_in,p3_in,p4_in,t1_in,t2_in,type_in,dim_in,sigma_in,scale_in,inv_in)
    integer                       :: ntapers
    integer, intent(in), optional :: type_in(ntapers),dim_in(ntapers)
    real   , intent(in), optional :: p1_in(ntapers),p2_in(ntapers),p3_in(ntapers),p4_in(ntapers)
    integer, intent(in), optional :: t1_in(3,ntapers),t2_in(3,ntapers)
    real,    intent(in), optional :: sigma_in(ntapers), scale_in(ntapers)
    logical, intent(in), optional :: inv_in(ntapers)

    integer                       :: type(ntapers),dim(ntapers),t1(3,ntapers),t2(3,ntapers)
    real                          :: p1(ntapers),p2(ntapers),p3(ntapers),p4(ntapers)
    real                          :: sigma(ntapers), scale(ntapers)
    logical                       :: inv(ntapers)

    integer                       :: ii

    if (present(p1_in)) then; p1 = p1_in; else;
      do ii=1,ntapers;        p1 = 1; end do; end if;
    if (present(p2_in)) then; p2 = p2_in; else;
      do ii=1,ntapers;        p2 = .5*pi; end do; end if;
    if (present(p3_in)) then; p3 = p3_in; else;
      do ii=1,ntapers;        p3 = .5*pi;; end do; end if;
    if (present(p4_in)) then; p4 = p4_in; else;
      do ii=1,ntapers;        p4 = .5*pi; end do; end if;
    if (present(t1_in)) then; t1 = t1_in; else;
      do ii=1,ntapers;        t1(:,ii) = (/ 1, 1, 1 /); end do; end if;
    if (present(t2_in)) then; t2 = t2_in; else;
      do ii=1,ntapers;        t2(:,ii) = n; end do; end if;
    if (present( type_in)) then;  type =  type_in; else;
      do ii=1,ntapers;  type(ii) = 0; end do; end if;
    if (present(sigma_in)) then; sigma = sigma_in; else;
      do ii=1,ntapers; sigma(ii) = 0; end do; end if;
    if (present(  inv_in)) then;   inv =   inv_in; else;
      do ii=1,ntapers;   inv(ii) = .false.; end do; end if;
    if (present(scale_in)) then; scale = scale_in; else;
      do ii=1,ntapers; scale(ii) = .false.; end do; end if;
    if (present(  dim_in)) then;   dim = dim_in; else;
      do ii=1,ntapers;   dim(ii) = 3;       end do; end if;
 
    do ii=1,ntapers
      if (p1(ii) .gt. p2(ii)) then; write(0,*) "Error - p2 < p1",ii,"-taper"; call exit(); end if
      if (p2(ii) .gt. p3(ii)) then; write(0,*) "Error - p3 < p2",ii,"-taper"; call exit(); end if
      if (p3(ii) .gt. p4(ii)) then; write(0,*) "Error - p3 < p2",ii,"-taper"; call exit(); end if
      call mdtapers_new_init( )
      call mdtapers_create  (p1_in=p1(ii),p2_in=p2(ii),p3_in=p3(ii),p4_in=p4(ii),type_in=type(ii),sigma_in=sigma(ii),scale_in=scale(ii),dim_in=dim(ii) )
      call mdtapers_insert  (.false., t1(:,ii), t2(:,ii),inv_in=inv_in(ii))
    end do

  end subroutine


  subroutine mdtapers_create(p1_in,p2_in,p3_in,p4_in,t1_in,t2_in,type_in,sigma_in,scale_in,dim_in)
    integer, intent(in),optional :: t1_in(3),t2_in(3),type_in,dim_in
    real   , intent(in),optional :: p1_in,p2_in,p3_in,p4_in
    real   , intent(in),optional :: sigma_in
    real   , intent(in),optional :: scale_in

    real                         :: p1,p2,p3,p4
    integer                      :: t1(3),t2(3)
    real                         :: sigma

    integer                      :: type,dim
    real                         :: scale

    if (present( p1_in)) then;  p1 = p1_in ; else;  p1 = 0    ; end if;
    if (present( p2_in)) then;  p2 = p2_in ; else;  p2 = .5*pi; end if;
    if (present( p3_in)) then;  p3 = p3_in ; else;  p3 = .5*pi; end if;
    if (present( p4_in)) then;  p4 = p4_in ; else;  p4 = .5*pi; end if;
    if (present( t1_in)) then;  t1 = t1_in ; else;  t1 = (/ 1, 1, 1 /)    ; end if;
    if (present( t2_in)) then;  t2 = t2_in ; else;  t2 = n;     end if;
    if (present(type_in )) then; type  = type_in;  else; type  = 1;   end if;
    if (present(dim_in  )) then; dim   = dim_in;   else; dim   = 3;   end if;
    if (present(sigma_in)) then; sigma = sigma_in; else; sigma = .4;  end if;
    if (present(scale_in)) then; scale = scale_in; else; scale = 1.;  end if;

  select case(type)
    
    case(1)
      call mdtapers_zero(                    )
    case(2)
      call mdtapers_rect  (.false.,p1,p2,t1,t2,scale_in=scale)
    case(3)
      call mdtapers_linear_up(.false.,p1,p2,t1,t2, scale_in=scale)
    case(4)
      call mdtapers_linear_down(.false.,p1,p2,t1,t2, scale_in=scale)
    case(5)
      call mdtapers_pyramid(.false.,p1,p2,t1,t2, scale_in=scale)
    case(6)
      call mdtapers_pyramid_block(.false., p1,p2,p3,p4,t1,t2,scale_in=scale)
    case(7)
      call mdtapers_hann(.false.,p1,p2,t1,t2, scale_in=scale)
    case(8)
      call mdtapers_hann_half(.false.,p1,p2,t1,t2, scale_in=scale)
    case(9)
      call mdtapers_hann_block(.false., p1,p2,p3,p4,t1,t2, scale_in=scale)
    case(10)
      call mdtapers_cosine   (.false.,p1,p2,t1,t2, scale_in=scale)
    case(11)
      call mdtapers_gaussian ( .false.,p1,p2,t1,t2, scale_in=scale, sigma_in=sigma )
    case(12)
      call mdtapers_gaussian_half(.false.,p1,p2,t1,t2, scale_in=scale, sigma_in=sigma )
    case(13)
      call mdtapers_gaussian_block(.false.,p1, p2, p3, p4,t1,t2, scale_in=scale, sigma_in=sigma )

  end select

  end subroutine


  ! Sum new 1d shape into 1d taper
  subroutine  mdtapers_compile(add,mdtaper_in,t1,t2)
    real   , intent(in)       :: mdtaper_in(:,:,:)
    logical, intent(in)       :: add
    integer                   :: t1(3),t2(3)
    integer                   :: ii,jj,kk
    
    call addnull(add,mdtaper_com( t1(1):t2(1),t1(2):t2(2),t1(3):t2(3) )  )
      mdtaper_com( t1(1):t2(1),t1(2):t2(2),t1(3):t2(3) ) = mdtaper_com( t1(1):t2(1),t1(2):t2(2),t1(3):t2(3) ) &
                                                         + mdtaper_in ( t1(1):t2(1),t1(2):t2(2),t1(3):t2(3) )

  end subroutine


  ! Invert a shape  x: (max(x)-x)
  subroutine mdtapers_invert(inv,x)
    real,intent(inout) :: x(:,:,:)
    logical,intent(in) :: inv    
    
    if (inv) then
      x = max(maxval(x),1.)-x
    end if
    
  end subroutine
  
  ! Get an indice given an absolute time, d and o
  subroutine mdtaper_indice(t,t_ind,d_in,o_in)
    real,optional,intent(in) :: d_in,o_in
    real                     :: d   ,o
    integer                  :: t_ind
    real                     :: t
    if (present(d_in)) then; d = d_in; else; d = 1.; end if
    if (present(o_in)) then; o = o_in; else; o = 1.; end if
    
    !write(0,*) "================"
    !write(0,*) int((t-o)/d+.5)+1
    !write(0,*) int((t-o)/d+.5)
    !write(0,*) int((t-o)/d)
    !write(0,*) "================"

    t_ind = int((t-o)/d+.5)+1
    
  end subroutine
  
  
  
  
  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  !     1D TAPER SHAPES, al but first, are angle based tapers
  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  
  subroutine mdtapers_zero(t1_in,t2_in) ! Atcually zeros a 3d cube
    integer, optional, intent(in)          :: t1_in(3),t2_in(3)
    integer                                :: t1(3),t2(3)

    if (present(t1_in)) then; t1 = t1_in; else; t1 = (/ 1, 1, 1 /); end if
    if (present(t2_in)) then; t2 = t2_in; else; t2 = n; end if
    ! To zero the taper in the memory of mod_taper
    ! Without deallocating it
    
    call addnull         (.false.,mdtaper)
    call mdtapers_compile(.false.,mdtaper,t1,t2)
    
  end subroutine
  
  
  
  subroutine mdtapers_rect(add,p1,p2,t1,t2,scale_in)
    logical, intent(in)          :: add
    real,    intent(in)          :: p1,p2
    integer, intent(in)          :: t1(3),t2(3)
    real, optional,intent(in)    :: scale_in

    real                         :: scale
    real                         :: pp,pmap(n(2),n(3))
    integer                      :: ii,jj,kk,pt2(2),pt3(2)

    call addnull(.false.,mdtaper)
    if (present(scale_in)) then; scale = scale_in; else; scale=1.; end if

    do jj=1,n(2);  do kk=1,n(3)
      pmap(jj,kk) = sqrt( (real(jj-c(2))*d(2))**2 + (real(kk-c(3))*d(3) )**2) + epsilon(p1)
    end do;        end do

    ! Construction of taper
    do ii=t1(1),t2(1)
      
      do jj=1,n(2);
        do kk=1,n(3);
          pp = atan( real(ii-c(1))*d(1) / pmap(jj,kk) )

          !write(0,*)real(ii-c(1))*d(1),  pmap(jj,kk),d(1),d(2),d(3)
          
          if (pp.ge.p1 .and. pp.le.p2) then
            mdtaper(ii,jj,kk) = 1.
          end if
          
        end do
      end do    
      
    end do

    call mdtapers_compile( add,scale*mdtaper,t1,t2 )
    
  end subroutine
  
  
  subroutine mdtapers_linear_up(add,p1,p2,t1,t2,scale_in)
    logical, intent(in)          :: add
    real,    intent(in)          :: p1,p2
    integer, intent(in)          :: t1(3),t2(3)
    real, optional,intent(in)    :: scale_in

    real                         :: scale
    real                         :: pp,pmap(n(2),n(3))
    integer                      :: ii,jj,kk,pt2(2),pt3(2)
    
    call addnull(.false.,mdtaper)
    if (present(scale_in)) then; scale = scale_in; else; scale=1.; end if
    
    do jj=1,n(2);  do kk=1,n(3)
      pmap(jj,kk) = (sqrt( (real(jj-c(2))*d(2))**2+(real(kk-c(3))*d(3) )**2)+epsilon(p1))
    end do;        end do

    ! Construction of taper
    do ii=t1(1),t2(1)
      
      do jj=1,n(2);
        do kk=1,n(3);
          pp = atan((real(ii-c(1))*d(1))/pmap(jj,kk))

          if (pp.ge.p1 .and. pp.le.p2) then
            mdtaper(ii,jj,kk) = real(pp-p1)/real(p2-p1)
          end if

        end do
      end do    
      
    end do

    call mdtapers_compile( add,scale*mdtaper,t1,t2 )
    
  end subroutine





  subroutine mdtapers_linear_down(add,p1,p2,t1,t2,scale_in)
    logical, intent(in)          :: add
    real,    intent(in)          :: p1,p2
    integer, intent(in)          :: t1(3),t2(3)
    real, optional,intent(in)    :: scale_in

    real                         :: scale
    real                         :: pp,pmap(n(2),n(3))
    integer                      :: ii,jj,kk,pt2(2),pt3(2)
    
    call addnull(.false.,mdtaper)
    if (present(scale_in)) then; scale = scale_in; else; scale=1.; end if
    
    do jj=1,n(2);  do kk=1,n(3)
      pmap(jj,kk) = (sqrt( (real(jj-c(2))*d(2))**2+(real(kk-c(3))*d(3) )**2)+epsilon(p1))
    end do;        end do

    ! Construction of taper
    do ii=t1(1),t2(1)
      
      do jj=1,n(2);
        do kk=1,n(3);
          pp = atan((real(ii-c(1))*d(1))/pmap(jj,kk))

          if (pp.ge.p1 .and. pp.le.p2) then
            mdtaper(ii,jj,kk) = 1.-real(pp-p1)/real(p2-p1)
          end if

        end do
      end do    
      
    end do

    call mdtapers_compile( add,scale*mdtaper,t1,t2 )
    
  end subroutine
  
  
  subroutine mdtapers_pyramid(add,p1,p2,t1,t2,scale_in)
    logical, intent(in)          :: add
    real,    intent(in)          :: p1,p2
    integer, intent(in)          :: t1(3),t2(3)
    real, optional,intent(in)    :: scale_in

    real                         :: scale
    real                         :: pp,pmap(n(2),n(3))
    integer                      :: ii,jj,kk,pt2(2),pt3(2)
    
    call addnull(.false.,mdtaper)
    if (present(scale_in)) then; scale = scale_in; else; scale=1.; end if
    
    do jj=1,n(2);  do kk=1,n(3)
      pmap(jj,kk) = (sqrt( (real(jj-c(2))*d(2))**2+(real(kk-c(3))*d(3) )**2)+epsilon(p1))
    end do;        end do

    ! Construction of taper
    do ii=t1(1),t2(1)

      do jj=1,n(2);
        do kk=1,n(3);
          pp = atan((real(ii-c(1))*d(1))/pmap(jj,kk))

          if (pp.ge.p1 .and. pp.le.p2) then
            mdtaper(ii,jj,kk) = 1.-real(abs(pp-real(p2+p1)/2.)) / (real(p2-p1)/2.)
          end if

        end do
      end do    
      
    end do

    call mdtapers_compile( add,scale*mdtaper,t1,t2 )
    
  end subroutine
  
  subroutine mdtapers_pyramid_block(add,p1,p2,p3,p4,t1,t2,scale_in)
    logical, intent(in)          :: add
    real,    intent(in)          :: p1,p2,p3,p4
    integer, intent(in)          :: t1(3),t2(3)
    real, optional,intent(in)    :: scale_in

    real                         :: scale
    real                         :: pp,pmap(n(2),n(3))
    integer                      :: ii,jj,kk,pt2(2),pt3(2)
    
    call addnull(.false.,mdtaper)
    if (present(scale_in)) then; scale = scale_in; else; scale=1.; end if
    
    do jj=1,n(2);  do kk=1,n(3)
      pmap(jj,kk) = (sqrt( (real(jj-c(2))*d(2))**2+(real(kk-c(3))*d(3) )**2)+epsilon(p1))
    end do;        end do

    ! Construction of taper
    do ii=t1(1),t2(1)
      
      do jj=1,n(2);
        do kk=1,n(3);
          pp = atan((real(ii-c(1))*d(1))/pmap(jj,kk))

          if (pp.ge.p1 .and. pp.le.p2) then
            mdtaper(ii,jj,kk) =    real(pp-p1)/real(p2-p1)
          end if
          if (pp.ge.p2 .and. pp.le.p3) then
            mdtaper(ii,jj,kk) = 1.
          end if
          if (pp.ge.p3 .and. pp.le.p4) then
            mdtaper(ii,jj,kk) = 1.-real(pp-p3)/real(p4-p3)
          end if

        end do
      end do  

    end do

    call mdtapers_compile( add,scale*mdtaper,t1,t2 )
    
  end subroutine
  
  
  subroutine mdtapers_hann(add,p1,p2,t1,t2,scale_in)
    logical, intent(in)          :: add
    real,    intent(in)          :: p1,p2
    integer, intent(in)          :: t1(3),t2(3)
    real, optional,intent(in)    :: scale_in

    real                         :: scale
    real                         :: pp,pmap(n(2),n(3))
    integer                      :: ii,jj,kk,pt2(2),pt3(2)
    
    call addnull(.false.,mdtaper)
    if (present(scale_in)) then; scale = scale_in; else; scale=1.; end if
    
    do jj=1,n(2);  do kk=1,n(3)
      pmap(jj,kk) = (sqrt( (real(jj-c(2))*d(2))**2+(real(kk-c(3))*d(3) )**2)+epsilon(p1))
    end do;        end do

    ! Construction of taper
    do ii=t1(1),t2(1)
      
      do jj=1,n(2);
        do kk=1,n(3);
          pp = atan((real(ii-c(1))*d(1))/pmap(jj,kk))

          if (pp.ge.p1 .and. pp.le.p2) then
            mdtaper(ii,jj,kk) = .5*(1.-cos(2.*pi*real(pp-p1)/real(p2-p1)))
          end if

        end do
      end do    
      
    end do
    
    call mdtapers_compile( add,scale*mdtaper,t1,t2 )
    
  end subroutine
  
  
  subroutine mdtapers_hann_half(add,p1,p2,t1,t2,scale_in)
    logical, intent(in)          :: add
    real,    intent(in)          :: p1,p2
    integer, intent(in)          :: t1(3),t2(3)
    real, optional,intent(in)    :: scale_in

    real                         :: scale
    real                         :: pp,pmap(n(2),n(3))
    integer                      :: ii,jj,kk,pt2(2),pt3(2)
    
    call addnull(.false.,mdtaper)
    if (present(scale_in)) then; scale = scale_in; else; scale=1.; end if
    
    do jj=1,n(2);  do kk=1,n(3)
      pmap(jj,kk) = (sqrt( (real(jj-c(2))*d(2))**2+(real(kk-c(3))*d(3) )**2)+epsilon(p1))
    end do;        end do

    ! Construction of taper
    do ii=t1(1),t2(1)
      
      do jj=1,n(2);
        do kk=1,n(3);
          pp = atan((real(ii-c(1))*d(1))/pmap(jj,kk))

          if (pp.ge.p1 .and. pp.le.p2) then
            mdtaper(ii,jj,kk) = .5*(1.-cos(pi*real(pp-p1)/real(p2-p1)))
          end if
          if (pp.ge.p2) then
            mdtaper(ii,jj,kk) = 1.
          end if

        end do
      end do    
      
    end do
    
    call mdtapers_compile( add,scale*mdtaper,t1,t2 )
    
  end subroutine
  
  
  subroutine mdtapers_hann_block (add,p1,p2,p3,p4,t1,t2,scale_in)
    logical, intent(in)          :: add
    real,    intent(in)          :: p1,p2,p3,p4
    integer, intent(in)          :: t1(3),t2(3)
    real, optional,intent(in)    :: scale_in

    real                         :: scale
    real                         :: pp,pmap(n(2),n(3))
    integer                      :: ii,jj,kk,pt2(2),pt3(2)
    
    call addnull(.false.,mdtaper)
    if (present(scale_in)) then; scale = scale_in; else; scale=1.; end if
    
    do kk=1,n(3);  do jj=1,n(2)
      pmap(jj,kk) = (sqrt( (real(jj-c(2))*d(2))**2+(real(kk-c(3))*d(3) )**2)+epsilon(p1))
    end do;        end do

    ! Construction of taper
    do ii=t1(1),t2(1)
      
      do jj=1,n(2);
        do kk=1,n(3);
          pp = atan((real(ii-c(1))*d(1))/pmap(jj,kk))

          if (pp.ge.p1 .and. pp.le.p2) then
            mdtaper(ii,jj,kk) = .5*( 1. - cos(pi*real(pp-p1)/real(p2-p1)) )
          end if
          if (pp.ge.p2 .and. pp.le.p3) then
            mdtaper(ii,jj,kk) = 1.
          end if
          if (pp.ge.p3 .and. pp.le.p4) then
            mdtaper(ii,jj,kk) = .5*( 1. + cos(pi*real(pp-p3)/real(p4-p3)) )
          end if

        end do
      end do    
      
    end do
    
    call mdtapers_compile( add,scale*mdtaper,t1,t2 )
    
  end subroutine
  
  
  subroutine mdtapers_cosine(add,p1,p2,t1,t2,scale_in)
    logical, intent(in)          :: add
    real,    intent(in)          :: p1,p2
    integer, intent(in)          :: t1(3),t2(3)
    real, optional,intent(in)    :: scale_in

    real                         :: scale
    real                         :: pp,pmap(n(2),n(3))
    integer                      :: ii,jj,kk,pt2(2),pt3(2)
    
    call addnull(.false.,mdtaper)
    if (present(scale_in)) then; scale = scale_in; else; scale=1.; end if
    
    do kk=1,n(3);  do jj=1,n(2)
      pmap(jj,kk) = (sqrt( (real(jj-c(2))*d(2))**2+(real(kk-c(3))*d(3) )**2)+epsilon(p1))
    end do;        end do

    ! Construction of taper
    do ii=t1(1),t2(1)
      
      do jj=1,n(2);
        do kk=1,n(3);
          pp = atan((real(ii-c(1))*d(1))/pmap(jj,kk))

          if (pp.ge.p1 .and. pp.le.p2) then
            mdtaper(ii,jj,kk) = sin(pi*(pp-p1)/(p2-p1))
          end if

        end do
      end do    
      
    end do

    call mdtapers_compile( add,scale*mdtaper,t1,t2 )
    
  end subroutine



  subroutine mdtapers_gaussian(add,p1,p2,t1,t2,scale_in,sigma_in)
    logical, intent(in)          :: add
    real,    intent(in)          :: p1,p2
    integer, intent(in)          :: t1(3),t2(3)
    real, optional,intent(in)    :: scale_in,sigma_in

    real                         :: scale,sigma
    real                         :: pp,pmap(n(2),n(3))
    integer                      :: ii,jj,kk,pt2(2),pt3(2)
    
    call addnull(.false.,mdtaper)
    if (present(scale_in)) then; scale = scale_in; else; scale=1.; end if
    if (present(sigma_in)) then; sigma = sigma_in; else; sigma=.4; end if
    
    do kk=1,n(3);  do jj=1,n(2)
      pmap(jj,kk) = (sqrt( (real(jj-c(2))*d(2))**2+(real(kk-c(3))*d(3) )**2)+epsilon(p1))
    end do;        end do

    ! Construction of taper
    do ii=t1(1),t2(1)

      do jj=1,n(2);
        do kk=1,n(3);
          pp = atan((real(ii-c(1))*d(1))/pmap(jj,kk))

          if (pp.ge.p1 .and. pp.le.p2) then
            mdtaper(ii,jj,kk) = exp(-.5*((((pp-p1)-((p2-p1)/2))/(sigma*((p2-p1)/2)))**2))
          end if

        end do
      end do    
      
    end do

    call mdtapers_compile( add,scale*mdtaper,t1,t2 )
    
  end subroutine



  subroutine mdtapers_gaussian_half(add,p1,p2,t1,t2,scale_in,sigma_in)
    logical, intent(in)          :: add
    real,    intent(in)          :: p1,p2
    integer, intent(in)          :: t1(3),t2(3)
    real, optional,intent(in)    :: scale_in,sigma_in

    real                         :: scale,sigma,pp2
    real                         :: pp,pmap(n(2),n(3))
    integer                      :: ii,jj,kk,pt2(2),pt3(2)
    
    call addnull(.false.,mdtaper)
    if (present(scale_in)) then; scale = scale_in; else; scale=1.; end if
    if (present(sigma_in)) then; sigma = sigma_in; else; sigma=.4; end if
    
    do kk=1,n(3);  do jj=1,n(2)
      pmap(jj,kk) = (sqrt( (real(jj-c(2))*d(2))**2+(real(kk-c(3))*d(3) )**2)+epsilon(p1))
    end do;        end do

    pp2 = p1 + 2*(p2-p1)
    ! Construction of taper
    do ii=t1(1),t2(1)

      do jj=1,n(2);
        do kk=1,n(3);
          pp = atan((real(ii-c(1))*d(1))/pmap(jj,kk))

          if (pp.ge.p1 .and. pp.le.p2) then
            mdtaper(ii,jj,kk) = exp(-.5*((((pp-p1)-((pp2-p1)/2))/(sigma*((pp2-p1)/2)))**2))
          end if
          if (pp.ge.p2) then
            mdtaper(ii,jj,kk) = 1.
          end if

        end do
      end do    
      
    end do

    call mdtapers_compile( add,scale*mdtaper,t1,t2 )
    
  end subroutine


  subroutine mdtapers_gaussian_block(add,p1,p2,p3,p4,t1,t2,scale_in,sigma_in)
    logical, intent(in)          :: add
    real,    intent(in)          :: p1,p2,p3,p4
    integer, intent(in)          :: t1(3),t2(3)
    real, optional,intent(in)    :: scale_in,sigma_in

    real                         :: scale,sigma,pp2,pp3
    real                         :: pp,pmap(n(2),n(3))
    integer                      :: ii,jj,kk,pt2(2),pt3(2)
    
    call addnull(.false.,mdtaper)
    if (present(scale_in)) then; scale = scale_in; else; scale=1.; end if
    if (present(sigma_in)) then; sigma = sigma_in; else; sigma=.4; end if
    
    do kk=1,n(3);  do jj=1,n(2)
      pmap(jj,kk) = (sqrt( (real(jj-c(2))*d(2))**2+(real(kk-c(3))*d(3) )**2)+epsilon(p1))
    end do;        end do

    pp2 = p1 + 2*(p2-p1)
    pp3 = p4 - 2*(p4-p3)
    ! Construction of taper
    do ii=t1(1),t2(1)

      do jj=1,n(2);
        do kk=1,n(3);
          pp = atan((real(ii-c(1))*d(1))/pmap(jj,kk))

          if (pp.ge.p1 .and. pp.le.p2) then
            mdtaper(ii,jj,kk) = exp(-.5*((((pp-p1)-((pp2-p1)/2))/(sigma*((pp2-p1)/2)))**2))
          end if
          if (pp.ge.p2 .and. pp.le.p3) then
            mdtaper(ii,jj,kk) = 1.
          end if
          if (pp.ge.p3 .and. pp.le.p4) then
            mdtaper(ii,jj,kk) = exp(-.5*((((pp-pp3)-((p4-pp3)/2))/(sigma*((p4-pp3)/2)))**2))
          end if

        end do
      end do    
      
    end do

    call mdtapers_compile( add,scale*mdtaper,t1,t2 )
    
  end subroutine



end module

