program srBandpass1d
  ! Program to bandpass a trace
  ! Sjoerd de Ridder, Stanford, 1/26/2012
  !
  ! Taper Specifics (like in srTaper1d)
  ! opt: (taper types 
  ! 0)  'zero'            --  needed: 'nothing' 
  ! 1)  'rect' (default)  --  needed: f1,f2
  ! 2)  'linear'          --  needed: f1,f2
  ! 3)  'pyramid'         --  needed: f1,f2
  ! 4)  'pyramid-band'    --  needed: f1,f2,f3,f4
  ! 5)  'hann'            --  needed: f1,f2
  ! 6)  'hann_half'       --  needed: f1,f2
  ! 7)  'hann-band'       --  needed: f1,f2,f3,f4
  ! 8)  'gaussian'        --  needed: f1,f2,sigma
  ! 9)  'gaussian_half'   --  needed: f1,f2,sigma
  ! 10) 'gaussian_block'  --  needed: f1,f2,f3,f4,sigma
  !
  ! f1,f2 default to spanning the axis.
  ! f1,f2,f3,f4 default to splitting the axis three way.
  !
  ! unit (for f1,f2,...):
  ! 1) Hz (default)
  ! 2) Radians
  !
  ! All tapers have the form of a band or lowcut,
  ! Notch tapers and high cut tapers are constructed
  ! with inv=1 option.
  ! Scale the data using parameter scale (default 1.). 
  ! Specify indices instead of physical units using f1_ind,f2_ind,...
  ! Inspect taper by supplying 'taper' as filename-parameter.
  !
  !
  ! How to treat the time axis (like in srSpectra1d):
  !
  ! true_t:
  ! T: use physical units padd to make symmetric*
  ! F: just stuff through FFT  (default)
  ! fftshift:
  ! T: fftshift data  (default)
  ! F: do not fftshift data
  ! 
  ! * use o1 and d1 from_param to redefine physical units
  !
  ! n2 needs to be divisible by 2 (for usr of r2c fftw)
  ! use true_t=1 to get around this (it padds one value).
  
  use sep                      ! Main SEP module
  implicit none
  
  ! Include for FFTW
  include "fftw3.f"

  ! Axis units
  integer                         :: n1,nt,nf,n2,n3,n4,n5,n6,b,e
  real                            :: d1,o1,dt,ot,df,of
  ! Memory variables
  real,allocatable, dimension(:)     :: datat,temp
  complex,allocatable, dimension(:)  :: dataf
  real,allocatable, dimension(:)     :: taper
  
  ! For Tapers
  integer                         :: f1_ind,f2_ind,f3_ind,f4_ind,hind
  real                            :: f1,    f2,    f3,    f4
  real                            :: scaler
  real                            :: sigma  ! For Gaussian taper
  integer                         :: opt
  logical                         :: inv
  integer                         :: funit    ! 1: Hz      2: Radians

  ! Misc
  integer                         :: ii,jj
  real, parameter                 :: pi=3.1415926535
  logical                         :: verbose,true_t,fftshift
  
  integer*8                       :: fft_for,fft_inv
  
  ! For geophone sensitivity
  real                            :: la,f0,sg,w0,f,w
  complex,allocatable             :: sensy(:)
  logical                         :: desensy
  
  ! Open I/O headerfiles
  call sep_init()
  call from_param('verbose',verbose,.false.)
  call from_param('verb'   ,verbose,verbose)
  if (verbose) write(0,*) '-------------------------------------------'
  if (verbose) write(0,*) '            BANDPASS A TRACE               '
  if (verbose) write(0,*) '-------------------------------------------'
  if (verbose) write(0,*) '        i/o looping over traces            '

  ! Read axis Details
  call from_history("n1",n1,1 )
  call from_history("n2",n2,1 )
  call from_history("n3",n3,1 )
  call from_history("n4",n4,1 )
  call from_history("n5",n5,1 )
  call from_history("n6",n6,1 )
  call from_history("d1",d1,1.)
  call from_history("o1",o1,0.)
  call from_param  ("o1",o1,o1)
  call from_param  ("d1",d1,d1)
  
  ! Can I do r2c FFTW?
  if (floor(real(n1)/2.).ne.ceiling(real(n1)/2.)) call erexit("This would be a lot easier if n1 is divisible by 2")
  
  call from_param('opt'  ,opt  ,1)  
  call from_param('funit',funit,1)
  
  call from_param('true_t'  ,true_t  ,.false.)  
  call from_param('fftshift',fftshift,.true. )
  
  ! true_t:
  if (verbose) then
    if ( opt.eq.1 ) write(0,*) "amplitude spectrum"
    if ( opt.eq.2 ) write(0,*) "power spectrum"
    write(0,*) "(fftshift,true_t)", (/fftshift,true_t/) 
    if (     fftshift) write(0,*) "fftshift data"
    if (.not.fftshift) write(0,*) "do not fftshift data"
    if (     true_t  ) write(0,*) "use physical units padd to make symmetric"
    if (.not.true_t  ) write(0,*) "just stuff through FFT"
  end if
  
  ! Determine how to treat time axis
  if (true_t) then
    ot = minval( (/o1,-abs(o1+n1*d1)/) )
    dt = d1
    nt = -ot/d1*2
    b  = (o1-ot)/d1+1
    e  = b+n1-1
  else
    ot=  0.
    dt= d1
    nt= n1
    b =  1
    e  = b+n1-1
  end if
  
  if (verbose) write(0,*) 'Defining FT axis as (-t,t): ',ot,ot+(nt-1)*d1
  if (verbose) write(0,*) 'read data into indices: ',b,e,'out of',nt
  
  ! Fourier units
  of = 0.
  df = 1./(nt*dt)
  nf = nt/2+1
  if (funit.eq.2) df=2.*pi*df
  
  ! Read input parameters for taper, there is lots of options :-)
  call from_param ("opt"  ,  opt,   1    )
  if (opt.lt.0..or.opt.gt.10) opt =  1
  call from_param ("inv"  , inv, .false. )
  call from_param ("sigma",sigma,   1.   )
  call from_param ("scale",scaler,  1.   )
  call from_param ("desensy",desensy,.true.)
  
  ! Read Tie points from param
  call from_param ("f1"    ,f1    , of          )
  call from_param ("f1_ind",f1_ind, nint((f1-of)/df)+1)
  
  if (opt.eq.4 .or. opt.eq.7 .or. opt.eq.10) then
  call from_param ("f2"    ,f2    , of+1*nf/3*df)
  call from_param ("f2_ind",f2_ind, nint((f2-of)/df)+1)
  else
  call from_param ("f2"    ,f2    , of+(nf-1)*df)
  call from_param ("f2_ind",f2_ind, nint((f2-of)/df)+1)
  end if
  
  call from_param ("f3"    ,f3    , of+2*nf/3*df)
  call from_param ("f3_ind",f3_ind, nint((f3-of)/df)+1)
  
  call from_param ("f4"    ,f4    , of+(nf-1)*df)
  call from_param ("f4_ind",f4_ind, nint((f4-of)/df)+1)

  ! Make sure they fall within range
  if (f1_ind.lt.1.) f1_ind=1
  if (f2_ind.lt.1.) f2_ind=1
  if (f3_ind.lt.1.) f3_ind=1
  if (f4_ind.lt.1.) f4_ind=1
  if (f1_ind.gt.n1) f1_ind=n1
  if (f2_ind.gt.n1) f2_ind=n1
  if (f3_ind.gt.n1) f3_ind=n1
  if (f4_ind.gt.n1) f4_ind=n1

  if (verbose) then
    if (opt.eq.4 .or. opt.eq.7 .or. opt.eq.10) then
      if (verbose) write(0,"(A,F,X,F,X,F,X,F)") ' taper tie values : ',f1    ,f2    ,f3    ,f4
      if (verbose) write(0,"(A,I3.3,X,I3.3,X,I3.3,X,I3.3)") ' taper tie indices: ',f1_ind,f2_ind,f3_ind,f4_ind
    else
      if (verbose) write(0,"(A,F,X,F)") ' taper tie values : ',f1    ,f2
      if (verbose) write(0,"(A,I,X,I)") ' taper tie indices: ',f1_ind,f2_ind
    end if
  end if
  
  
  !!! TAPER CREATION
  allocate(taper(nf)); taper=0.
  select case(opt)

    case(0)
      if (verbose) write(0,*) 'taper: 0.'

    case(1)
      if (verbose) write(0,*) 'taper: rectangle'
      taper(f1_ind:f2_ind)=1.

    case(2)
      if (verbose) write(0,*) 'taper: linear'
      taper(1:f1_ind)=1.
      do ii=f1_ind,f2_ind
        taper(ii)=1.-real(ii-f1_ind)/real(f2_ind-f1_ind)
      end do
      
    case(3)
      if (verbose) write(0,*) 'taper: pyramid'
      hind = nint(real(f2_ind-f1_ind)/2.)
      do ii=f1_ind,f1_ind+hind
        taper(ii)=   real(ii-f1_ind     )/real(f1_ind+hind-f1_ind)
      end do
      do ii=f1_ind+hind,f2_ind
        taper(ii)=1.-real(ii-f1_ind-hind)/real(f2_ind-f1_ind-hind)
      end do
      
    case(4)
      if (verbose) write(0,*) 'taper: pyramid-band'
      hind = nint(real(f2_ind-f1_ind)/2.)
      do ii=f1_ind,f2_ind
        taper(ii)=   real(ii-f1_ind     )/real(f2_ind-f1_ind)
      end do
      taper(f2_ind:f3_ind) = 1.
      do ii=f3_ind,f4_ind
        taper(ii)=1.-real(ii-f3_ind     )/real(f4_ind-f3_ind)
      end do
      
    case(5)
      if (verbose) write(0,*) 'taper: hann'
      do ii=f1_ind,f2_ind
        taper(ii)= .5*(1.-cos(2.*pi*(ii-f1_ind)/real(f2_ind-f1_ind)))
      end do
      
    case(6)
      if (verbose) write(0,*) 'taper: hann-half'
      taper(1:f1_ind) = 1.
      do ii=f1_ind,f2_ind
        taper(ii)= .5*(1.+cos(   pi*(ii-f1_ind)/real(f2_ind-f1_ind)))
      end do
      
    case(7)
      if (verbose) write(0,*) 'taper: hann-band'
      do ii=f1_ind,f2_ind
        taper(ii)= .5*(1.-cos(pi*(ii-f1_ind)/real(f2_ind-f1_ind)))
      end do
      taper(f2_ind:f3_ind) = 1.
      do ii=f3_ind,f4_ind
        taper(ii)= .5*(1.+cos(pi*(ii-f3_ind)/real(f4_ind-f3_ind)))
      end do

    case(8)
      if (verbose) write(0,*) 'taper: gaussian'
      hind = nint(real(f2_ind-f1_ind))
      do ii=f1_ind,f2_ind
        taper(ii)= exp(-(ii-hind)**2./(2*d1*sigma**2.) )
      end do
      
    case(9)
      if (verbose) write(0,*) 'taper: gaussian-half'
      taper(1:f1_ind)=1.
      do ii=f1_ind,f2_ind
        taper(ii)= exp(-((ii-f1_ind)**2.)/(2*d1*sigma**2.) )
      end do
      
    case(10)
      if (verbose) write(0,*) 'taper: gaussian-band'
      do ii=f1_ind,f2_ind
        taper(ii)= exp(-((ii-f2_ind)**2.)/(2*d1*sigma**2.) )
      end do
      taper(f2_ind:f3_ind) = 1.
      do ii=f3_ind,f4_ind
        taper(ii)= exp(-((ii-f3_ind)**2.)/(2*d1*sigma**2.) )
      end do
  
  end select
  
  ! Invert taper to lowpass or notch
  if (inv) taper = 1.-taper
  
  ! Taper as output
  if (exist_file("taper")) then
    if (verbose) write(0,*) 'write taper to file'
    call to_history("n1", nf ,"taper")
    call to_history("d1", df ,"taper")
    call to_history("o1", of ,"taper")
    call srite("taper",taper,4*nf)
  end if
  
  ! Close all I/O to headerfiles
  call sep_close ()
  
  ! Prepare geophone sensitivity
  la = 68
  sg = .220*100
  f0 = 15
  w0 = 2*pi*f0
  sensy = 0.
  allocate(sensy(nf))
  do ii=1,nf
    f  = of+(ii-1)*df
    w  = 2*pi*f
    sensy(ii) = sg*(w**2./w0**2.)/((-w**2./w0**2.)+cmplx(1,2*la/100.*w/w0))
  end do
  
  ! Allocate memories for trace
  allocate(datat(nt),dataf(nf),temp(nt/2))
  ! Prepare fftw plans
  call sfftw_plan_dft_r2c_1d(fft_for, nt,  datat, dataf, FFTW_ESTIMATE);
  call sfftw_plan_dft_c2r_1d(fft_inv, nt,  dataf, datat, FFTW_ESTIMATE);
  ! Use taper to implement fft scaling
  taper = taper / real(n1)
  
  
  if (verbose) write(0,*) '-------------------------------------------'
  !!!! Actual Kernel
  do jj=1,n2*n3*n4*n5*n6
    if (verbose) write(0,"(A,I7.1,A,I7.3)") 'trace: ',jj,' out of',n2*n3*n4*n5*n6
    
    datat = 0.; dataf=0.
    ! Read the data
    call sreed('in',datat(b:e),4*n1)
    
    ! fftshift the data
    if (fftshift) then
      temp               = datat(     1:nt/2)
      datat(     1:nt/2) = datat(nt/2+1:nt  )
      datat(nt/2+1:nt  ) = temp
    end if
    
    ! fft the data
    call sfftw_execute( fft_for )

    ! Optionally remove the Geophone response    
    if (desensy) then
     dataf    = dataf/1000./sensy
     dataf(1) = 0
    end if
    
    ! Actual taper the spectrum (all fft scaling is done with taper)
    dataf = taper*dataf
    
    
    ! ifft the data
    call sfftw_execute( fft_inv )
    
    ! ifftshift the data
    if (fftshift) then
      temp               = datat(nt/2+1:nt  )
      datat(nt/2+1:nt  ) = datat(     1:nt/2)
      datat(     1:nt/2) = temp
    end if
    
    ! Write the data
    call srite('out',datat(b:e),4*n1)
    
  end do
  
  
  ! Deallocate Memory
  if (verbose) write(0,*) '-------------------------------------------'
  call sfftw_destroy_plan(fft_for,fft_inv)
  deallocate(datat,dataf,temp)
  deallocate(sensy)
  call exit(0)

  
end program

