program srSparseSelect
  ! Program to select a geographic subportion of a dataset
  ! and get rid of empty traces.
  !
  ! xy_in, xy_out : header files in and out
  ! 
  ! Modify and add selection options as you see fit
  ! 
  ! opt:
  ! 0)  selects just everything (except when sparse=T*)
  ! 1)  selects for absolute distance from cx and cy
  !     needs xy_in
  !     r(=0)  ! distance in header units (default returns very little)
  !     xind,yind(=1,2) ! indices in header file of x and coordinate
  !     cx,cy ! center of geopgraphic subportion in header units
  !     unit(=1.)  ! scale the header units
  ! 
  ! Additional Options:
  ! sparse: (T*/F) dont select empty traces
  ! nkill=0     number of traces to mute
  ! kill=i,i,i  indices of traces to mute
  
  
  use sep
  implicit none
  
  integer                            :: n1  ,n2  ,n3,  n4,  n5,  n6
  integer                            :: n1xy,n2xy,n3xy,n4xy,n5xy,n6xy
  integer                            :: ntraces,ii,jj,nkill
  real, allocatable, dimension(:)    :: datain
  real, allocatable, dimension(:)    :: header
  integer, allocatable, dimension(:) :: kill
  real                               :: r,cx,cy,unit
  integer                            :: xind,yind
  logical                            :: keep,sparse,sparse2,verbose
  integer                            :: opt
  
  ! Open I/O headerfiles 
  call sep_init()
  call from_param('verbose',verbose,.false.)
  call from_param('verb'   ,verbose,verbose)
  if (verbose) write(0,*) '-----------------------------------'
  if (verbose) write(0,*) '      SELECT GEOGRAPHIC SUBSET     '
  if (verbose) write(0,*) '-----------------------------------'
  if (verbose) write(0,*) '    i/o looping over traces        '

  call from_history('n1',n1,1)
  call from_history('n2',n2,1)
  call from_history('n3',n3,1)
  call from_history('n4',n4,1)
  call from_history('n5',n5,1)
  call from_history('n6',n6,1)
  
  if (exist_file('xy_in')) then
    call from_aux('xy_in','n1',n1xy,1)
    call from_aux('xy_in','n2',n2xy,1)
    call from_aux('xy_in','n3',n3xy,1)
    call from_aux('xy_in','n4',n4xy,1)
    call from_aux('xy_in','n5',n5xy,1)
    call from_aux('xy_in','n6',n6xy,1)
  
    if (n2xy.ne.n2) call erexit('n2xy.ne.n2')
    if (n3xy.ne.n3) call erexit('n3xy.ne.n3')
    if (n4xy.ne.n4) call erexit('n4xy.ne.n4')
    if (n5xy.ne.n5) call erexit('n5xy.ne.n5')
    if (n6xy.ne.n6) call erexit('n6xy.ne.n6')
  end if

  ! Selection criteria:
  call from_param('sparse'  ,sparse  ,.true. )
  call from_param('sparse2' ,sparse2 ,.false.)
  if (sparse2) sparse=.true.
  
  call from_param('opt'     ,opt     ,0      )
  select case (opt)
   case(1)
     if (.not.exist_file('xy_in')) call erexit("Need xy_in for opt=1")
     call from_param('r' ,r ,1.)  ! Distance (in unit of header units)
     call from_param('cx',cx,0.)  ! Center_x (in unit of header units)
     call from_param('cy',cy,0.)  ! Center_y (in unit of header units)
  end select

  ! Usefull for all selection options
  call from_param('xind',xind,1 )
  call from_param('yind',yind,2 )
  call from_param('unit',unit,1.)

  ! Mute traces  
  call from_param('nkill',nkill,0)
  allocate(kill(nkill))
  if (nkill.ge.1) call from_param('kill',kill  ,(/1/))

  allocate(datain(n1))
  if (exist_file('xy_in')) allocate(header(n1xy))
  

  if (verbose) write(0,*) '-------------------------------------------'
  !!!! Actual Kernel
  ntraces=0
  do jj=1,n2*n3*n4*n5*n6
    keep=.true.
    header=0.
    datain=0.

    ! Get header for sure
    if (exist_file('xy_in')) call sreed("xy_in",header,4*n1xy)

    ! Mute or not
    do ii=1,nkill
      if (jj.eq.kill(ii)) keep=.false.
    end do

    ! Options check
    select case (opt)
     case(1)
       if (sqrt((header(xind)*unit-cx)**2.+(header(yind)*unit-cy)**2.)>r) keep=.false.
    end select


    if (keep) then
      ! Read data
      call sreed_window("in",2,(/n1,n2*n3*n4*n5*n6/),(/n1,1/),(/0,(jj-1)/),(/1,1/),4,datain)
      if ( sum(abs(datain(nint(real(n1)/10):nint(real(n1)/10*9)))).eq.0. ) then
!        write(0,*) 'zeroed'
        datain=0.
      end if

      ! Sparse check
      if (sparse) then
        if (sum(abs(datain)).eq.0. .or. sum(abs(datain)).gt.huge(r)) keep=.false.
      end if
    end if

    ! Verbose the Diagnosis
    if (verbose) write(0,"(A,I7.1,A,I7.3,A,L,F)") 'trace: ',jj,' out of',n2*n3*n4*n5*n6,' keep: ', keep

    if (keep) then ! Write data
      call srite("out"   ,datain,4*n1  )
      if (exist_file('xy_out')) call srite("xy_out",header,4*n1xy)
      ntraces = ntraces+1
    end if
    if (.not.keep .and. sparse2) then
      call srite("out"   ,datain,4*n1  )
 !     write(0,*) 'still kept'
      header=0.
      if (exist_file('xy_out')) call srite("xy_out",header,4*n1xy)
      ntraces = ntraces+1
    end if
    
  end do
  
  deallocate(datain)
  if (exist_file('xy_in')) deallocate(header)
  call to_history('n2',ntraces)
  call to_history('n3',1)
  call to_history('n4',1)
  call to_history('n5',1)
  call to_history('n6',1)
  if (exist_file('xy_out')) then
    call to_history('n1',n1xy   ,'xy_out')
    call to_history('n2',ntraces,'xy_out')
    call to_history('n3',1,'xy_out')
    call to_history('n4',1,'xy_out')
    call to_history('n5',1,'xy_out')
    call to_history('n6',1,'xy_out')
  end if
  call sep_close()
  call exit(0)
  
end program
