program srSpectra1d_desenty
  ! Program to compute Fourier spectrum (uses FFTW)
  ! Sjoerd de Ridder, Stanford, 1/26/2012
  !
  ! opt:
  ! 1) amplitude spectrum (default)
  ! 2) power spectrum
  !
  ! unit:
  ! 1) Hz (default)
  ! 2) Radians
  !
  ! true_t:
  ! T: use physical units padd to make symmetric*
  ! F: just stuff through FFT  (default)
  ! fftshift:
  ! T: fftshift data  (default)
  ! F: do not fftshift data
  ! 
  ! * use o1 and d1 from_param to redefine physical units
  !
  ! n2 needs to be divisible by 2 (for usr of r2c fftw)
  ! use true_t=1 to get around this (it padds one value).
  
  use sep
  implicit none
  
  ! Include for FFTW
  include "fftw3.f"

  ! Memory variables
  real, allocatable, dimension(:)     :: datat,temp
  complex,allocatable, dimension(:)   :: dataf
  
  ! Axis units
  integer                         :: n1,nt,nf,n2,n3,n4,n5,n6,b,e
  real                            :: d1,o1,dt,ot,df,of
  
  ! Spectrum type
  logical                         :: power    ! power spectrum or amsplitude spectrum
  integer                         :: funit    ! 1: Hz      2: Radians
  real                            :: vref
  logical                         :: db
  
  ! Misc
  integer                         :: ii,jj,kk,ll,mm
  real, parameter                 :: pi=3.1415926535
  logical                         :: verbose,true_t,fftshift
  
  integer*8                       :: fftplan
  
  ! For geophone sensitivity
  real                            :: la,f0,sg,w0,f,w
  complex,allocatable             :: sensy(:)

  real,allocatable                :: spamp(:)
  double precision                :: mean
  
  
  ! Open I/O headerfiles 
  call sep_init()
  call from_param('verbose',verbose,.false.)
  call from_param('verb'   ,verbose,verbose)
  if (verbose) write(0,*) '-----------------------------------'
  if (verbose) write(0,*) '        COMPUTE 1D SPECTRA         '
  if (verbose) write(0,*) '-----------------------------------'
  if (verbose) write(0,*) '    i/o looping over traces        '
  
  ! Read from History
  call from_history("n1",n1,1 )
  call from_history("n2",n2,1 )
  call from_history("n3",n3,1 )
  call from_history("n4",n4,1 )
  call from_history("n5",n5,1 )
  call from_history("n6",n6,1 )
  call from_history("o1",o1,0.)
  call from_history("d1",d1,1.)
  call from_param  ("o1",o1,o1)
  call from_param  ("d1",d1,d1)
  
  ! Can I do r2c FFTW?
  if (floor(real(n1)/2.).ne.ceiling(real(n1)/2.)) call erexit("This would be a lot easier if n1 is divisible by 2")
  
  call from_param('power',power,.false.)  
  call from_param('funit',funit,1)

  call from_param('true_t'  ,true_t  ,.false.)  
  call from_param('fftshift',fftshift,.true. )
  
  ! true_t:
  if (verbose) then
    if ( .not. power ) write(0,*) "amplitude spectrum"
    if (       power ) write(0,*) "power spectrum"
    write(0,*) "(fftshift,true_t)", (/fftshift,true_t/) 
    if (     fftshift) write(0,*) "fftshift data"
    if (.not.fftshift) write(0,*) "do not fftshift data"
    if (     true_t  ) write(0,*) "use physical units padd to make symmetric"
    if (.not.true_t  ) write(0,*) "just stuff through FFT"
  end if

  ! Determine how to treat time axis
  if (true_t) then
    ot = minval( (/o1,-abs(o1+n1*d1)/) )
    dt = d1
    nt = -ot/d1*2
    b  = (o1-ot)/d1+1
    e  = b+n1-1
  else
    ot=  0.
    dt= d1
    nt= n1
    b =  1
    e  = b+n1-1
  end if

  if (verbose) write(0,*) 'Defining FT axis as (-t,t): ',ot,ot+(nt-1)*d1
  if (verbose) write(0,*) 'read data into indices: ',b,e,'out of',nt
  
  ! Fourier units
  of = 0.
  df = 1./(nt*dt)
  nf = nt/2+1
  if (funit.eq.2) df=2.*pi*df

  ! Write new History
  call to_history("n1",nf)
  call to_history("d1",df)
  call to_history("o1",of)

  call from_param('db',db,.true.)
  call from_param('vref',vref,.00000005)

  call sep_close()

  ! Allocate memories for trace
  allocate(datat(nt),dataf(nf),temp(nt/2),sensy(nf),spamp(nf))
  ! Initialize fftw plan  
  call sfftw_plan_dft_r2c_1d(fftplan, nt,  datat, dataf, FFTW_ESTIMATE);

  ! Prepare geophone sensitivity
  la = 68
  sg = .220*100.  ! V/(cm/s) -> V/(m/s)
  f0 = 15
  w0 = 2*pi*f0
  sensy = 0.
  do ii=1,nf
    f  = (ii-1)*df
    w  = 2*pi*f
    sensy(ii) = sg*(w**2./w0**2.)/((-w**2./w0**2.)+cmplx(1,2*la/100.*w/w0))
  end do



  if (verbose) write(0,*) '-------------------------------------------'
  !!!! Actual Kernel
  do jj=1,n2*n3*n4*n5*n6
    if (verbose) write(0,"(A,I7.1,A,I7.3)") 'trace: ',jj,' out of',n2*n3*n4*n5*n6
    
    datat = 0.; dataf=0.; spamp=0.
    ! Read the data
    call sreed('in',datat(b:e),4*n1)
    datat(b:e) = datat(b:e)/1000.

 !   write(0,*) b,e
 !   write(0,*) datat(1:10)

    mean = sum(dble(datat(b:e)))/dble(n1)
    !write(0,*) mean

    datat(b:e) = real( dble(datat(b:e))-mean)

 !   mean = sum(dble(datat(b:e)))/dble(n1)
 !   write(0,*) mean
    
  !  write(0,*) b,e
  !  write(0,*) datat(1:10)

    ! fftshift the data
    if (fftshift) then
      temp               = datat(     1:nt/2)
      datat(     1:nt/2) = datat(nt/2+1:nt  )
      datat(nt/2+1:nt  ) = temp
    end if

    
    ! fft the data
    call sfftw_execute( fftplan )
    !write(0,*) 'dataf(1:10)'
    !write(0,*) dataf(1:10)*dt
    dataf = dt*dataf

    ! Compute and scale the spectrum
    do ii=1,nf
      spamp(ii) = cabs(dataf(ii)) / cabs(sensy(ii)) !V / V/(m/s) = m/s
    end do

    ! power or amplitude:
    if ( .not. power ) then
      do ii=1,nf
        spamp(ii) = 2*spamp(ii)*df
      end do
    else
      do ii=1,nf
        spamp(ii) = 2*spamp(ii)*spamp(ii)*df*df
      end do
    end if

    ! write(0,*) 'spamp'
    !  write(0,*) spamp(1:10)
    
    if ( db ) then
      if ( .not. power ) then
        do ii=1,nf
          spamp(ii) = 20*log10(spamp(ii)/vref)
        end do
      else
        do ii=1,nf
          spamp(ii) = 10*log10(spamp(ii)/vref)
        end do
      end if
    end if
    ! write(0,*) 'spamp'
    !  write(0,*) spamp(1:10)
    where (spamp.ne.spamp) spamp=0.
    where (abs(spamp).gt.huge(spamp)) spamp=0.
    
    ! Write the data
    call srite('out',spamp,4*nf)
    
  end do

  ! Deallocate memory
  if (verbose) write(0,*) '-----------------------------------'
  call sfftw_destroy_plan(fftplan)
  deallocate(datat,dataf,temp)
  call exit(0)
  
end program
