module leaky_helix_mod

  ! TO DO: Include the subroutines from the following modules inside this module's source code
  use helix
  use helicon
  use polydiv
  implicit none

  type (filter), private    :: filt
  integer      , private    :: stat
  logical      , private    :: adj_fop


  interface leaky_fop
    module procedure   leaky_1d_fop
    module procedure   leaky_2d_fop
    module procedure   leaky_3d_fop
  end interface

  interface leaky_int_lop
    module procedure   leaky_int_1d_lop
    module procedure   leaky_int_2d_lop
    module procedure   leaky_int_3d_lop
  end interface

  interface leaky_dif_lop
    module procedure   leaky_dif_1d_lop
    module procedure   leaky_dif_2d_lop
    module procedure   leaky_dif_3d_lop
  end interface

  contains


  subroutine leaky_init   (rho,n,adj)
    real, intent(in)    :: rho
    integer,intent(in)  ::     n
    logical,intent(in),optional :: adj

    call allocatehelix(filt,1)
    filt%lag(1) =    1
    filt%flt(1) = -rho
    call helicon_init(filt)    ! For Forward
    call polydiv_init(n,filt)  ! For Inverse

    if (present(adj)) then; adj_fop = adj; else; adj_fop = .false.; end if

  end subroutine

  subroutine leaky_adjoint(adj)
    logical,intent(in) :: adj
    adj_fop = adj
  end subroutine
  
  
  subroutine leaky_close( )
    call deallocatehelix(filt)
    call helicon_close( )
    call polydiv_close( )
  end subroutine


  subroutine leaky_1d_fop    (inv,add,x,y)
    logical,intent(in)  :: inv,add
    real                ::          x(:),y(:)
    
    if (inv) then
      stat=polydiv_lop(adj_fop,add,y,x)
    else
      stat=helicon_lop(adj_fop,add,x,y)
    end if
    
  end subroutine

  subroutine leaky_2d_fop    (inv,add,x,y)
    logical,intent(in)  :: inv,add
    real                ::          x(:,:),y(:,:)
    integer             :: jj,n2

    do jj=1,n2
      if (inv) then
        stat=polydiv_lop( adj_fop, add, y(:,jj), x(:,jj) )
      else
        stat=helicon_lop( adj_fop, add, x(:,jj), y(:,jj) )
      end if
    end do
  
  end subroutine


  subroutine leaky_3d_fop    (inv,add,x,y)
    logical,intent(in)  :: inv,add
    real                ::          x(:,:,:),y(:,:,:)
    integer             :: jj,n2,kk,n3

    do kk=1,n3
    do jj=1,n2
      if (inv) then
        stat=polydiv_lop( adj_fop, add, y(:,jj,kk), x(:,jj,kk) )
      else
        stat=helicon_lop( adj_fop, add, x(:,jj,kk), y(:,jj,kk) )
      end if
    end do
    end do
  
  end subroutine




  subroutine leaky_int_init (rho,n)
    real, intent(in)    :: rho
    integer,intent(in)  ::     n
    call allocatehelix(filt,1)
    filt%lag(1) =    1
    filt%flt(1) = -rho
    call polydiv_init(n,filt)
  end subroutine
  
  subroutine leaky_int_close( )
    call deallocatehelix(filt)
    call polydiv_close( )
  end subroutine


  subroutine leaky_int_1d_lop(adj,add,x,y)
    logical,intent(in)  :: adj,add
    real                ::         x(:),y(:)
    
    stat=polydiv_lop(adj,add,y,x)

  end subroutine


  subroutine leaky_int_2d_lop(adj,add,x,y)
    logical,intent(in)  :: adj,add
    real                ::         x(:,:),y(:,:)
    integer             :: jj,n2
    
    do jj=1,n2
      stat=polydiv_lop(adj,add,x(:,jj),y(:,jj))
    end do
    
  end subroutine


  subroutine leaky_int_3d_lop(adj,add,x,y)
    logical,intent(in)  :: adj,add
    real                ::         x(:,:,:),y(:,:,:)
    integer             :: jj,n2,kk,n3
    
    do kk=1,n3
    do jj=1,n2
      stat=polydiv_lop(adj,add,y(:,jj,kk),x(:,jj,kk))
    end do
    end do
    
  end subroutine
  
  
  subroutine leaky_dif_init (rho)
    real, intent(in)    :: rho
    call allocatehelix(filt,1)
    filt%lag(1) =    1
    filt%flt(1) = -rho
    call helicon_init(filt)
  end subroutine
  
  
  subroutine leaky_dif_close( )
    call deallocatehelix(filt)
    call helicon_close( )
  end subroutine
  
  
  subroutine leaky_dif_1d_lop(adj,add,x,y)
    logical,intent(in)  :: adj,add
    real                ::         x(:),y(:)
    
    stat=helicon_lop(adj,add,y,x)
    
  end subroutine


  subroutine leaky_dif_2d_lop(adj,add,x,y)
    logical,intent(in)  :: adj,add
    real                ::         x(:,:),y(:,:)
    integer             :: jj,n2

    do jj=1,n2
      stat=helicon_lop(adj,add,y(:,jj),x(:,jj))
    end do

  end subroutine


  subroutine leaky_dif_3d_lop(adj,add,x,y)
    logical,intent(in)  :: adj,add
    real                ::         x(:,:,:),y(:,:,:)
    integer             :: jj,n2,kk,n3
    
    do kk=1,n3
    do jj=1,n2
      stat=helicon_lop(adj,add,y(:,jj,kk),x(:,jj,kk))
    end do
    end do
    
  end subroutine


end module
