module mtspectrum_mod
  ! Contains my (optimized) versions of German's MTSPECTRA functions

  use dpss_mod
  use mt_fftw_mod

  implicit none

!  Memory variables of MT FFT's
   complex,private,allocatable                            :: dk(:,:)

!  Dpss 
   real   ,private                                        :: tbp, df
   integer, private                                       :: nt,n2,nf,nk
   real(8),private, allocatable                           :: lambda(:), theta(:),vn(:,:)

!  Error level of adaptive estimation
   logical,private                                        :: adapt
   real(8),private                                        :: rerr

  contains
  

subroutine mtspec_init      (nt_in, n2_in, nf_in, df_in, tbp_in, nk_in,adapt_in,rerr_in)
  real,intent(in)         ::                      df_in, tbp_in
  integer, intent(in)     :: nt_in, n2_in, nf_in,                nk_in
  logical, optional       ::                                           adapt_in         
  real(8), optional       ::                                                    rerr_in
  
  nt  = nt_in
  n2  = n2_in
  nf  = nf_in
  df  = df_in
  nk  = nk_in
  tbp = tbp_in
  
  if ( present(adapt_in) ) then; adapt = adapt_in; else; adapt = .false.; end if;
  if (          nk == 1  ) then;                         adapt = .false.; end if; 
  if ( present(rerr_in ) ) then; rerr  = rerr_in;  else; rerr  =  9.5d-7; end if;
  
  allocate(lambda(nk),theta(nk),vn(nt,nk),dk(nf,nk))

  ! Calculating tapers
  if (nt<20000) then
     call dpss(             nt/2,dble(tbp),nk,vn,lambda,theta)
  else
     !  write(0,*) 'Computing DPSS with interpolation' 
     call dpss_spline(10000,nt/2,dble(tbp),nk,vn,lambda,theta)
  end if
  
  call mt_fftw_r2c_lop_init(nt,nf,nk,real(vn))
  
end subroutine


subroutine mtspec_close( )
  
  deallocate( lambda,theta,vn,dk )
  call mt_fftw_r2c_lop_close( )
  
end subroutine


subroutine mtspec (m,d)
  ! model is temporal-domain; data is fourier-domain.
   real(4)    , intent(in)                          :: m(:,:)
   real(4)    , intent(inout)                       :: d(:,:)
   
   integer              :: ii, jj, kk
   
   do jj = 1,n2
     write(0,*) 'Trace:',jj,'/',n2
     
     ! Compute MT FTT
     call mt_fftw_r2c_lop(.false.,.false.,m(:,jj),dk)
     
     ! Calculate adaptive weigthed spectrum
     call adaptspec(dk,d(:,jj))
     
   end do
   
end subroutine


subroutine adaptspec(yk,d)

   real(4)       :: d(:)
   complex(4)    :: yk  (: ,: )
   complex(8)    :: yk_d(nf,nk)

   
   !  Iterations
   real(8)       :: wt(nf,nk)
   real(8)       :: slast(nf), sbar(nf)
   real(8)       :: skw(nf,nk)

   !  Other
   integer       :: jj, kk
   real(8)       :: dvar
   real(8)       :: bk(nk)   , varsk(nk)
   real(8)       :: evalu(nk), sqev(nk)


   if ( .not. adapt ) then

     d = sum(abs(yk),2)/nk

   else

    !  Initialize some variables
     yk_d = dble(abs(yk))

     bk = sum( ( yk_d(1,:) + yk_d(nf,:) + 2.d0*sum(yk_d(2:nf-1,:),1) ) * dble( df ) )/dble(nk) * theta

     evalu = 1.d0 - theta
     sqev  = evalu !sqrt(evalu)

  !
  !  step through frequencies and find weighted spectrum iteratively for each frequency
  !
     sbar = ( yk_d(:,1) + yk_d(:,2) )/2.d0
     do jj = 1,1000

        slast = sbar

        do kk = 1,nk
           wt (:,kk) = sqev(kk)*sbar /( evalu(kk)*sbar + bk(kk))
           wt (:,kk) = min(wt(:,kk),1.d0)
           skw(:,kk) = abs(wt(:,kk)) * yk_d(:,kk)  
        end do
        
        sbar = sum(skw,2)/sum(abs(wt),2)
	
        if (jj==1000) then
            d = real(sbar)
            ! write(0,*) 'adaptspec did not converge, rerr = ', maxval(abs((sbar-slast)/(sbar+slast))),rerr
           cycle
        endif
        if (maxval(abs((sbar-slast)/(sbar+slast))) .gt. rerr) then
           cycle
        endif
        !write(0,*) 'finished',jj
        d = real(sbar)
        exit
        
     end do

   end if
      
end subroutine


end module
