module slepian_mod

  implicit none
  contains
 


subroutine dpss(npts, fw, nev, v, lambda, theta)

!
!  Modified:
!	German Prieto
!	November 2004
!
!  Calculation of the Discrete Prolate Spheroidal Sequences, and 
!  the correspondent eigenvalues. Also, the (1 - eigenvalue) terms
!  are calculated. 
!
!  Slepian, D.     1978  Bell Sys Tech J v57 n5 1371-1430
!  Thomson, D. J.  1982  Proc IEEE v70 n9 1055-1096
!
!  Input:
!    npts    	the number of points in the series
!    fw    	the time-bandwidth product (number of Rayleigh bins)
!    nev   	the desired number of tapers
!
!  Output:
!    v     	the eigenvectors (tapers) are returned in v(npts,nev)
!    lambda 	the eigenvalues of the v's
!    theta	the 1-lambda values. The energy outside the bandwidth
!
!  The tapers are the eigenvectors of the tridiagonal matrix sigma(i,j)
!  [see Slepian(1978) eq 14 and 25.] They are also the eigenvectors of
!  the Toeplitz matrix eq. 18. We solve the tridiagonal system in
!  tridib and tinvit for the tapers and use them in the integral 
!  equation in the frequency domain (dpss_ev subroutine) to get the
!  eigenvalues more accurately, by performing Chebychev Gaussian 
!  Quadrature following Thomson's codes.
!
!  First, we create the three diagonals of the tridiagonal matrix. We
!  compute separetely the even and odd tapers, by calling tridib 
!  (eigenvalues) and tinvit (eigenfunctions), from eispack. 
!  We, refine the eigenvalues, by computing the inner bandwidth 
!  energy in the frequency domain (eq. 2.6 Thomson). Also the "leakage"
!  (1 - eigenvalue) is estimated, independenly if necesary. 
!
!  calls tridib, tinvit, dpss_ev
!

!********************************************************************

   implicit none

   real(8), parameter :: pi=3.141592653589793d0, r2=1.4142135623731d0

   integer :: ntot, lh, nr, k, kr, k2, m11, n, i, ierr, nx
   
   real(8) :: atol, sn, eps1, lb, ub

!  Important variables

   integer, intent(in) :: npts, nev
   real(8), intent(in) :: fw 

   real(8), intent(out), dimension(nev) :: lambda, theta
   real(8), intent(out), dimension(npts,nev) :: v

!  Working variables

   integer :: neven, nodd
   integer, dimension(nev) :: ind
   
   real(8) :: bw, hn, com
   real(8), dimension(nev) :: eigval
   real(8), dimension(:), allocatable :: fv1, fv2, fv3
   real(8), dimension(:,:), allocatable :: v2

!********************************************************************

   if (npts < 2) then
      return
   endif

   nx=mod(npts,2)
   lh=(npts/2)+nx

   nodd  = nev/2
   neven = nev - nodd

   bw = fw/dble(npts)
   com = cos(2.d0*pi*bw)

   hn = dble(npts-1)/2.d0

   if (allocated(fv1)) then
      deallocate(fv1)
   endif 
   if (allocated(fv2)) then
      deallocate(fv2)
   endif 
   if (allocated(fv3)) then
      deallocate(fv3)
   endif 

   allocate(fv1(lh))
   allocate(fv2(lh))
   allocate(fv3(lh))

!
!  Perform symmetry reduction to half size
!

   do i=1,lh
      n = i-1

!     Main diagonal
      
      fv1(i)    = com*(hn - dble(n))**2.d0

!     sub diagonal

      fv2(i)    = dble(n*(npts-n))/2.d0
      fv3(i)    = fv2(i) * fv2(i)
   enddo
   
   if (nx.eq.0) then
      fv1(lh)   = com* (hn - dble(lh-1))**2.d0 + dble(lh*(npts-lh))/2.d0
   else
      fv2(lh)   = r2*fv2(lh)
      fv3(lh)   = 2.d0*fv3(lh)
   endif
  
!
!  Do the even tapers
!
   
   if (allocated(v2)) then
      deallocate(v2)
   endif   
   allocate(v2(lh,neven))
   
 
   eps1 = 0.d0
   m11 = lh-neven+1
   
   call tridib(lh,eps1,fv1,fv2,fv3,lb,ub,m11,neven,eigval,ind,ierr)   
   if (ierr .ne. 0) then
      write(6,'(a,i5)') 'tridib error ',ierr
      stop
   endif

   call tinvit(lh,fv1,fv2,fv3,neven,eigval,ind,v2,ierr)
   if (ierr .ne. 0) then
      write(6,'(a,i5)') 'tinvit error ',ierr
      stop
   endif


   if (nx==1) then
      do k = 1,neven
         v2(lh,k) = r2*v2(lh,k)
      enddo
   endif

   do k = 1,neven
      kr = neven - k + 1
      k2 = 2*k - 1

      theta(k2) = eigval(kr)

!
! Expand the eigenfunctions
!

      nr=npts
      do i=1,lh
         v(i,k2) = v2(i,kr)
         v(nr,k2)= v2(i,kr)
         nr=nr-1
      enddo

!
! Normalize the eigenfunction
!

      sn=0.d0
      do n=1,npts
         sn=sn+v(n,k2)*v(n,k2)
      enddo
      
      sn=1.d0/sqrt(sn)

!
! Put eigenfunctions positive standard
!

      if ((v(lh+1,k2).lt.0.d0)) then
         sn=-sn
      endif
      
      do n=1,npts
         v(n,k2)=sn*v(n,k2)
      enddo
  
   enddo

!
!  Do the odd tapers
!

   if (nodd > 0) then   

      if (allocated(v2)) then
         deallocate(v2)
      endif   
      allocate(v2(lh-nx,nodd))

      do i=1,lh
         n = i-1
         fv1(i)  = com*(hn - dble(n))**2
         fv2(i)  = dble(n*(npts-n))/2.d0
         fv3(i)  = fv2(i) * fv2(i)
      enddo
   
      if (nx.eq.0) then
         fv1(lh)  =  com* (hn - dble(lh-1))**2 - dble(lh*(npts-lh))/2.d0
      endif
   
      eps1 = 0.d0
      m11 = (lh-nx)-nodd+1
   
      call tridib(lh-nx,eps1,fv1,fv2,fv3,lb,ub,m11,nodd,eigval,ind,ierr)   
      if (ierr .ne. 0) then
         write(6,'(a,i5)') 'tridib error ',ierr
         stop
      endif
   
      call tinvit(lh-nx,fv1,fv2,fv3,nodd,eigval,ind,v2,ierr)
      if (ierr .ne. 0) then
         write(6,'(a,i5)')'tinvit error ',ierr
         stop
      endif

      
      do k = 1,nodd
         kr = nodd - k + 1
         k2 = 2*k 

         theta(k2) = eigval(kr)

!
! Expand the eigenfunctions
!

         nr=npts
         do i=1,lh-nx
            v(i,k2) = v2(i,kr)
            v(nr,k2)= -v2(i,kr)
            nr=nr-1
         enddo
         if (nx == 1) then
            v(lh,k2) = 0.d0
         endif
   
!
! Normalize the eigenfunction
!

         sn=0.d0
         do n=1,npts
            sn=sn+v(n,k2)*v(n,k2)
         enddo
      
         sn=1.d0/sqrt(sn)

!
! Put eigenfunctions positive standard
!

         if ((nx==1) .and. (v(lh+1,k2).lt.0.d0)) then
            sn = -sn    
         endif
         if ((nx <= 0) .and. (v(lh+1,k2).lt.0.d0)) then
            sn=-sn
         endif

         do n=1,npts
            v(n,k2)=sn*v(n,k2)
         enddo

      enddo

   endif

   ntot=neven+nodd

!
!  Get the eigenvalues, by Quadrature (Chebychev)
!

   atol = 1.d-14

   call dpss_ev(npts,nev,bw,atol,v,lambda,theta)

   deallocate(fv1,fv2,fv3,v2)


   return
   
end subroutine dpss


subroutine dpss_spline(nmax, npts, fw, nev, v_sp, lambda, theta)

!
!  Modified:
!	German Prieto
!	July 2005
!
!  Calculation of the Discrete Prolate Spheroidal Sequences, and 
!  the correspondent eigenvalues. Also, the (1 - eigenvalue) terms
!  are calculated. 
!
!  Slepian, D.     1978  Bell Sys Tech J v57 n5 1371-1430
!  Thomson, D. J.  1982  Proc IEEE v70 n9 1055-1096
!
!  Input:
!    nmax	the number of actual points to calculate dpss
!    npts    	the number of points in the series
!    fw    	the time-bandwidth product (number of Rayleigh bins)
!    nev   	the desired number of tapers
!
!  Output:
!    v_sp     	the eigenvectors (tapers) are returned in v(npts,nev)
!    lambda 	the eigenvalues of the v's
!    theta	the 1-lambda values. The energy outside the bandwidth
!
!  The tapers are the eigenvectors of the tridiagonal matrix sigma(i,j)
!  [see Slepian(1978) eq 14 and 25.] They are also the eigenvectors of
!  the Toeplitz matrix eq. 18. We solve the tridiagonal system in
!  tridib and tinvit for the tapers and use them in the integral 
!  equation in the frequency domain (dpss_ev subroutine) to get the
!  eigenvalues more accurately, by performing Chebychev Gaussian 
!  Quadrature following Thomson's codes.
!
!  First, we create the three diagonals of the tridiagonal matrix. We
!  compute separetely the even and odd tapers, by calling tridib 
!  (eigenvalues) and tinvit (eigenfunctions), from eispack. 
!  We, refine the eigenvalues, by computing the inner bandwidth 
!  energy in the frequency domain (eq. 2.6 Thomson). Also the "leakage"
!  (1 - eigenvalue) is estimated, independenly if necesary. 
!
!  calls tridib, tinvit, dpss_ev
!

!********************************************************************

   implicit none

   real(8), parameter :: pi=3.141592653589793d0, r2=1.4142135623731d0

   integer :: ntot, lh, nr, k, kr, k2, m11, n, i, ierr, nx
   
   real(8) :: atol, sn, eps1, lb, ub

!  Important variables

   integer, intent(in) :: npts, nev, nmax
   real(8), intent(in) :: fw 

   real(8), intent(out), dimension(nev) :: lambda, theta
   real(8), intent(out), dimension(npts,nev) :: v_sp

!  Working variables

   integer :: neven, nodd
   integer, dimension(nev) :: ind
   
   real(8) :: bw, hn, com
   real(8), dimension(nev) :: eigval
   real(8), dimension(:), allocatable :: fv1, fv2, fv3
   real(8), dimension(:,:), allocatable :: v2

!  Interpolation

   integer :: j

   real(8), dimension(nmax) :: t, u, ddu 

   real(8), dimension(nmax,nev) :: v

   real(8) :: dt, tx, x2, xddu, xdu

!********************************************************************


   if (npts < 2) then
      return
   endif

   nx=mod(nmax,2)
   lh=(nmax/2)+nx

   nodd  = nev/2
   neven = nev - nodd

   bw = fw/dble(nmax)
   com = cos(2.d0*pi*bw)

   hn = dble(nmax-1)/2.d0

   if (allocated(fv1)) then
      deallocate(fv1)
   endif 
   if (allocated(fv2)) then
      deallocate(fv2)
   endif 
   if (allocated(fv3)) then
      deallocate(fv3)
   endif 

   allocate(fv1(lh))
   allocate(fv2(lh))
   allocate(fv3(lh))

!
!  Perform symmetry reduction to half size
!

   do i=1,lh
      n = i-1

!     Main diagonal
      
      fv1(i)    = com*(hn - dble(n))**2.d0

!     sub diagonal

      fv2(i)    = dble(n*(nmax-n))/2.d0
      fv3(i)    = fv2(i) * fv2(i)
   enddo
   
   if (nx.eq.0) then
      fv1(lh)   = com* (hn - dble(lh-1))**2.d0 + dble(lh*(nmax-lh))/2.d0
   else
      fv2(lh)   = r2*fv2(lh)
      fv3(lh)   = 2.d0*fv3(lh)
   endif
  
!
!  Do the even tapers
!
   
   if (allocated(v2)) then
      deallocate(v2)
   endif   
   allocate(v2(lh,neven))
   
 
   eps1 = 0.d0
   m11 = lh-neven+1
   
   call tridib(lh,eps1,fv1,fv2,fv3,lb,ub,m11,neven,eigval,ind,ierr)   
   if (ierr .ne. 0) then
      write(6,'(a,i5)') 'tridib error ',ierr
      stop
   endif

   call tinvit(lh,fv1,fv2,fv3,neven,eigval,ind,v2,ierr)
   if (ierr .ne. 0) then
      write(6,'(a,i5)') 'tinvit error ',ierr
      stop
   endif


   if (nx==1) then
      do k = 1,neven
         v2(lh,k) = r2*v2(lh,k)
      enddo
   endif

   do k = 1,neven
      kr = neven - k + 1
      k2 = 2*k - 1

      theta(k2) = eigval(kr)

!
! Expand the eigenfunctions
!

      nr=nmax
      do i=1,lh
         v(i,k2) = v2(i,kr)
         v(nr,k2)= v2(i,kr)
         nr=nr-1
      enddo

!
! Normalize the eigenfunction
!

      sn=0.d0
      do n=1,nmax
         sn=sn+v(n,k2)*v(n,k2)
      enddo
      
      sn=1.d0/sqrt(sn)

!
! Put eigenfunctions positive standard
!

      if ((v(lh+1,k2).lt.0.d0)) then
         sn=-sn
      endif
      
      do n=1,nmax
         v(n,k2)=sn*v(n,k2)
      enddo
  
   enddo

!
!  Do the odd tapers
!

   if (nodd > 0) then   

      if (allocated(v2)) then
         deallocate(v2)
      endif   
      allocate(v2(lh-nx,nodd))

      do i=1,lh
         n = i-1
         fv1(i)  = com*(hn - dble(n))**2
         fv2(i)  = dble(n*(nmax-n))/2.d0
         fv3(i)  = fv2(i) * fv2(i)
      enddo
   
      if (nx.eq.0) then
         fv1(lh)  =  com* (hn - dble(lh-1))**2 - dble(lh*(nmax-lh))/2.d0
      endif
   
      eps1 = 0.d0
      m11 = (lh-nx)-nodd+1
   
      call tridib(lh-nx,eps1,fv1,fv2,fv3,lb,ub,m11,nodd,eigval,ind,ierr)   
      if (ierr .ne. 0) then
         write(6,'(a,i5)') 'tridib error ',ierr
         stop
      endif
   
      call tinvit(lh-nx,fv1,fv2,fv3,nodd,eigval,ind,v2,ierr)
      if (ierr .ne. 0) then
         write(6,'(a,i5)')'tinvit error ',ierr
         stop
      endif

      
      do k = 1,nodd
         kr = nodd - k + 1
         k2 = 2*k 

         theta(k2) = eigval(kr)

!
! Expand the eigenfunctions
!

         nr=nmax
         do i=1,lh-nx
            v(i,k2) = v2(i,kr)
            v(nr,k2)= -v2(i,kr)
            nr=nr-1
         enddo
         if (nx == 1) then
            v(lh,k2) = 0.d0
         endif
   
!
! Normalize the eigenfunction
!

         sn=0.d0
         do n=1,nmax
            sn=sn+v(n,k2)*v(n,k2)
         enddo
      
         sn=1.d0/sqrt(sn)

!
! Put eigenfunctions positive standard
!

         if ((nx==1) .and. (v(lh+1,k2).lt.0.d0)) then
            sn = -sn    
         endif
         if ((nx <= 0) .and. (v(lh+1,k2).lt.0.d0)) then
            sn=-sn
         endif

         do n=1,nmax
            v(n,k2)=sn*v(n,k2)
         enddo

      enddo

   endif

   ntot=neven+nodd

!  Spline interpolation

   do i = 1,nmax
      t(i) = dble(i)
   enddo
   
   dt = (t(nmax)-t(1))/dble(npts-1)

   do i = 1,nev
      
      u(1:nmax) = v(:,i)

      call spline_cubic_set ( nmax, t, u, 0, 0.d0, 0, 0.d0, ddu )


      do j=1, npts
         tx=t(1) + dble(j-1)*dt
         call spline_cubic_val(nmax,t,u,ddu,tx,x2,xdu,xddu)
         v_sp(j,i) = x2
      enddo

   enddo

!  Again, normalize interpolated eigenfunction

   do i = 1,nev

      sn=0.d0
      do j=1,npts
         sn=sn+v_sp(j,i)*v_sp(j,i)
      enddo
   
      sn=1.d0/sqrt(sn)

      do j=1,npts
         v_sp(j,i)=sn*v_sp(j,i)
      enddo

   enddo


!
!  Get the eigenvalues, by Quadrature (Chebychev)
!

   bw = fw/dble(npts) 

   atol = 1.d-14

   call dpss_ev(npts,nev,bw,atol,v_sp,lambda,theta)

   deallocate(fv1,fv2,fv3,v2)
   return
   
end subroutine dpss_spline



subroutine dpss_ev(ndata,nefn,w,atol,efn,evalu,theta)

!
!  Modified
!	German Prieto
!	November 2004
!
!   computes eigenvalues for the discrete prolate spheroidal sequences
!   in efn by integration of the corresponding squared discrete prolate
!   spheroidal wavefunctions over the inner domain. Due to symmetry, we
!   perform integration from zero to w. 
!   We use Chebychev quadrature for the numerical integration. 
! 
!   Input:
!
!      ndata  - actual number of time points in each column of efn
!      nefn   - number of dpss (= number of columns in efn)
!      w      - the bandwidth (= time-bandwidth product/ndata)
!      atol   - absolute error tolerance for the integration. this should
!               be set to 10**-n, where n is the number of significant figures
!               that can be be represented on the machine.
!      efn    - array containing the dpss
!
!   Output:
!
!      evalu  - vector of length nefn to contain the eigenvalues
!      theta - vector of length nefn to contain (1-eigenvalues)
! 
! 
!   calls functions xint
! 

!**********************************************************************

   implicit none

   
   integer, intent(in) :: ndata, nefn
   real(8), intent(in) :: atol,w
   real(8), dimension(ndata,nefn), intent(in) :: efn
   real(8), intent(out), dimension(nefn) :: evalu, theta

   integer :: k,i
   real(8) :: rup, rlow, sum
   
   real(8) :: result

 !  real(8) :: xint

!**********************************************************************
   
   do k=1,nefn
     

      result = xint(0.d0,w,atol,efn(1,k),ndata)

      
      evalu(k) = 2.d0*result
  
! 
!   calculate 1-lambda, explicitly if necessary
!

      if ((1.0d0-evalu(k)) .le. 1d-13) then

! 
!        break up 1-evalu calculation into reasonable pieces 
!        so xint can converge
! 

         sum=0.d0


         do i = 1,floor(.5d0/w)
            rlow = dble(i)*w
            if (rlow >= .5d0) then
               exit
            endif
            rup  = min(rlow+w,.5d0)
            result = xint(rlow,rup,atol,efn(1,k),ndata)
            sum=sum+result
         enddo

	 theta(k)=2.d0*sum
        
      else
      
         theta(k)=1.0d0-evalu(k)
      
      endif
     
   enddo

   return
      
end subroutine dpss_ev








!--------------------------------------------------------------------
! Spline subroutines
!--------------------------------------------------------------------

subroutine spline_cubic_set ( n, t, y, ibcbeg, ybcbeg, ibcend, ybcend, ypp )

!
! SPLINE_CUBIC_SET computes the second derivatives of a piecewise cubic spline.
!
!  Discussion:
!
!    For data interpolation, the user must call SPLINE_CUBIC_SET to 
!    determine the second derivative data, passing in the data to be 
!    interpolated, and the desired boundary conditions.
!
!    The data to be interpolated, plus the SPLINE_CUBIC_SET output, 
!    defines the spline.  The user may then call SPLINE_CUBIC_VAL to 
!    evaluate the spline at any point.
!
!    The cubic spline is a piecewise cubic polynomial.  The intervals
!    are determined by the "knots" or abscissas of the data to be
!    interpolated.  The cubic spline has continous first and second
!    derivatives over the entire interval of interpolation.  
!
!    For any point T in the interval T(IVAL), T(IVAL+1), the form of
!    the spline is
!
!      SPL(T) = A(IVAL)
!             + B(IVAL) * ( T - T(IVAL) ) 
!             + C(IVAL) * ( T - T(IVAL) )**2
!             + D(IVAL) * ( T - T(IVAL) )**3
!
!    If we assume that we know the values Y(*) and YPP(*), which represent
!    the values and second derivatives of the spline at each knot, then
!    the coefficients can be computed as:
!
!      A(IVAL) = Y(IVAL)
!      B(IVAL) = ( Y(IVAL+1) - Y(IVAL) ) / ( T(IVAL+1) - T(IVAL) )
!        - ( YPP(IVAL+1) + 2 * YPP(IVAL) ) * ( T(IVAL+1) - T(IVAL) ) / 6
!      C(IVAL) = YPP(IVAL) / 2
!      D(IVAL) = ( YPP(IVAL+1) - YPP(IVAL) ) / ( 6 * ( T(IVAL+1) - T(IVAL) ) )
!
!    Since the first derivative of the spline is
!
!      SPL'(T) =     B(IVAL)
!              + 2 * C(IVAL) * ( T - T(IVAL) )
!              + 3 * D(IVAL) * ( T - T(IVAL) )**2,
!
!    the requirement that the first derivative be continuous at interior
!    knot I results in a total of N-2 equations, of the form:
!
!      B(IVAL-1) + 2 C(IVAL-1) * (T(IVAL)-T(IVAL-1)) 
!      + 3 * D(IVAL-1) * (T(IVAL) - T(IVAL-1))**2 = B(IVAL)
!
!    or, setting H(IVAL) = T(IVAL+1) - T(IVAL)
!
!      ( Y(IVAL) - Y(IVAL-1) ) / H(IVAL-1)
!      - ( YPP(IVAL) + 2 * YPP(IVAL-1) ) * H(IVAL-1) / 6
!      + YPP(IVAL-1) * H(IVAL-1)
!      + ( YPP(IVAL) - YPP(IVAL-1) ) * H(IVAL-1) / 2
!      = 
!      ( Y(IVAL+1) - Y(IVAL) ) / H(IVAL)
!      - ( YPP(IVAL+1) + 2 * YPP(IVAL) ) * H(IVAL) / 6
!
!    or
!
!      YPP(IVAL-1) * H(IVAL-1) + 2 * YPP(IVAL) * ( H(IVAL-1) + H(IVAL) )
!      + YPP(IVAL) * H(IVAL) 
!      =
!      6 * ( Y(IVAL+1) - Y(IVAL) ) / H(IVAL)
!      - 6 * ( Y(IVAL) - Y(IVAL-1) ) / H(IVAL-1)    
!
!    Boundary conditions must be applied at the first and last knots.  
!    The resulting tridiagonal system can be solved for the YPP values.
!
!  Modified:
!
!    27 January 2004
!
!  Author:
!
!    John Burkardt
!
!  Reference:
!
!    Carl de Boor,
!    A Practical Guide to Splines,
!    Springer Verlag, 1978.
!
!  Parameters:
!
!    Input, integer N, the number of data points; N must be at least 2. 
!
!    Input, real ( kind = 8 ) T(N), the points where data is specified.  
!    The values should be distinct, and increasing.
!
!    Input, real ( kind = 8 ) Y(N), the data values to be interpolated.
!
!    Input, integer IBCBEG, the left boundary condition flag:
!
!      0: the spline should be a quadratic over the first interval;
!      1: the first derivative at the left endpoint should be YBCBEG;
!      2: the second derivative at the left endpoint should be YBCBEG.
!
!    Input, real ( kind = 8 ) YBCBEG, the left boundary value, if needed.
!
!    Input, integer IBCEND, the right boundary condition flag:
!
!      0: the spline should be a quadratic over the last interval;
!      1: the first derivative at the right endpoint should be YBCEND;
!      2: the second derivative at the right endpoint should be YBCEND.
!
!    Input, real ( kind = 8 ) YBCEND, the right boundary value, if needed.
!
!    Output, real ( kind = 8 ) YPP(N), the second derivatives of 
!    the cubic spline.
!

!********************************************************************

  implicit none

  integer :: n

  real(8), dimension(3,n) :: a
  integer :: i, ibcbeg, ibcend
  real(8), dimension(n) :: t, y, ypp
  real(8) :: ybcbeg, ybcend

!********************************************************************

!
!  Check.
!
  if ( n <= 1 ) then
    write ( *, '(a)' ) ' '
    write ( *, '(a)' ) 'SPLINE_CUBIC_SET - Fatal error!'
    write ( *, '(a)' ) '  The number of knots must be at least 2.'
    write ( *, '(a,i6)' ) '  The input value of N = ', n
    stop
  end if

  do i = 1, n-1
    if ( t(i+1) <= t(i) ) then
      write ( *, '(a)' ) ' '
      write ( *, '(a)' ) 'SPLINE_CUBIC_SET - Fatal error!'
      write ( *, '(a)' ) '  The knots must be strictly increasing, but'
      write ( *, '(a,i6,a,g14.6)' ) '  T(',  i,') = ', t(i)
      write ( *, '(a,i6,a,g14.6)' ) '  T(',i+1,') = ', t(i+1)
      stop
    end if
  end do
!
!  Set the first equation.
!
  if ( ibcbeg == 0 ) then
    ypp(1) = 0.0D+00
    a(2,1) = 1.0D+00
    a(1,2) = -1.0D+00
  else if ( ibcbeg == 1 ) then
    ypp(1) = ( y(2) - y(1) ) / ( t(2) - t(1) ) - ybcbeg
    a(2,1) = ( t(2) - t(1) ) / 3.0D+00 
    a(1,2) = ( t(2) - t(1) ) / 6.0D+00
  else if ( ibcbeg == 2 ) then
    ypp(1) = ybcbeg
    a(2,1) = 1.0D+00
    a(1,2) = 0.0D+00
  else
    write ( *, '(a)' ) ' '
    write ( *, '(a)' ) 'SPLINE_CUBIC_SET - Fatal error!'
    write ( *, '(a)' ) '  The boundary flag IBCBEG must be 0, 1 or 2.'
    write ( *, '(a,i6)' ) '  The input value is IBCBEG = ', ibcbeg
    stop
  end if
!
!  Set the intermediate equations.
!
  do i = 2, n-1
    ypp(i) = ( y(i+1) - y(i) ) / ( t(i+1) - t(i) ) &
           - ( y(i) - y(i-1) ) / ( t(i) - t(i-1) )
    a(3,i-1) = ( t(i) - t(i-1) ) / 6.0D+00
    a(2,i) = ( t(i+1) - t(i-1) ) / 3.0D+00
    a(1,i+1) = ( t(i+1) - t(i) ) / 6.0D+00
  end do
!
!  Set the last equation.
!
  if ( ibcend == 0 ) then
    ypp(n) = 0.0D+00
    a(3,n-1) = -1.0D+00
    a(2,n) = 1.0D+00
  else if ( ibcend == 1 ) then
    ypp(n) = ybcend - ( y(n) - y(n-1) ) / ( t(n) - t(n-1) )
    a(3,n-1) = ( t(n) - t(n-1) ) / 6.0D+00
    a(2,n) = ( t(n) - t(n-1) ) / 3.0D+00
  else if ( ibcend == 2 ) then
    ypp(n) = ybcend
    a(3,n-1) = 0.0D+00
    a(2,n) = 1.0D+00
  else
    write ( *, '(a)' ) ' '
    write ( *, '(a)' ) 'SPLINE_CUBIC_SET - Fatal error!'
    write ( *, '(a)' ) '  The boundary flag IBCEND must be 0, 1 or 2.'
    write ( *, '(a,i6)' ) '  The input value is IBCEND = ', ibcend
    stop
  end if
!
!  Special case:
!    N = 2, IBCBEG = IBCEND = 0.
!
  if ( n == 2 .and. ibcbeg == 0 .and. ibcend == 0 ) then

    ypp(1) = 0.0D+00
    ypp(2) = 0.0D+00
!
!  Solve the linear system.
!
  else

    call d3_np_fs ( n, a, ypp, ypp )

  end if

  return
end subroutine

!--------------------------------------------------------------------

subroutine spline_cubic_val ( n, t, y, ypp, tval, yval, ypval, yppval )

!
! SPLINE_CUBIC_VAL evaluates a piecewise cubic spline at a point.
!
!  Discussion:
!
!    SPLINE_CUBIC_SET must have already been called to define the 
!    values of YPP.
!
!    For any point T in the interval T(IVAL), T(IVAL+1), the form of
!    the spline is
!
!      SPL(T) = A 
!             + B * ( T - T(IVAL) ) 
!             + C * ( T - T(IVAL) )**2
!             + D * ( T - T(IVAL) )**3
!
!    Here:
!      A = Y(IVAL)
!      B = ( Y(IVAL+1) - Y(IVAL) ) / ( T(IVAL+1) - T(IVAL) )
!        - ( YPP(IVAL+1) + 2 * YPP(IVAL) ) * ( T(IVAL+1) - T(IVAL) ) / 6
!      C = YPP(IVAL) / 2
!      D = ( YPP(IVAL+1) - YPP(IVAL) ) / ( 6 * ( T(IVAL+1) - T(IVAL) ) )
!
!  Modified:
!
!    20 November 2000
!
!  Author:
!
!    John Burkardt
!
!  Reference:
!
!    Carl de Boor,
!    A Practical Guide to Splines,
!    Springer Verlag, 1978.
!
!  Parameters:
!
!    Input, integer N, the number of data values.
!
!    Input, real ( kind = 8 ) T(N), the knot values.
!
!    Input, real ( kind = 8 ) Y(N), the data values at the knots.
!
!    Input, real ( kind = 8 ) YPP(N), the second derivatives of the 
!    spline at the knots.
!
!    Input, real ( kind = 8 ) TVAL, a point, typically between T(1) and 
!    T(N), at which the spline is to be evalulated.  If TVAL lies outside 
!    this range, extrapolation is used.
!
!    Output, real ( kind = 8 ) YVAL, YPVAL, YPPVAL, the value of the spline, and
!    its first two derivatives at TVAL.
!

!********************************************************************

  implicit none

  integer n

  real (8) :: dt, h, tval, yppval, ypval, yval

  integer :: left, right

  real (8 ), dimension(n) :: t, y, ypp

!********************************************************************

!
!  Determine the interval [T(LEFT), T(RIGHT)] that contains TVAL.
!  Values below T(1) or above T(N) use extrapolation.
!
  call dvec_bracket ( n, t, tval, left, right )
!
!  Evaluate the polynomial.
!
  dt = tval - t(left)
  h = t(right) - t(left)

  yval = y(left) &
       + dt * ( ( y(right) - y(left) ) / h &
              - ( ypp(right) / 6.0D+00 + ypp(left) / 3.0D+00 ) * h &
       + dt * ( 0.5D+00 * ypp(left) &
       + dt * ( ( ypp(right) - ypp(left) ) / ( 6.0D+00 * h ) ) ) )

  ypval = ( y(right) - y(left) ) / h &
       - ( ypp(right) / 6.0D+00 + ypp(left) / 3.0D+00 ) * h &
       + dt * ( ypp(left) &
       + dt * ( 0.5D+00 * ( ypp(right) - ypp(left) ) / h ) )

  yppval = ypp(left) + dt * ( ypp(right) - ypp(left) ) / h 

  return
end subroutine

!--------------------------------------------------------------------

subroutine spline_cubic_val2 ( n, t, y, ypp, left, tval, yval, ypval, yppval )

!
! SPLINE_CUBIC_VAL2 evaluates a piecewise cubic spline at a point.
!
!  Discussion:
!
!    This routine is a modification of SPLINE_CUBIC_VAL; it allows the
!    user to speed up the code by suggesting the appropriate T interval
!    to search first.
!
!    SPLINE_CUBIC_SET must have already been called to define the
!    values of YPP.
!
!    In the LEFT interval, let RIGHT = LEFT+1.  The form of the spline is
!
!      SPL(T) = 
!          A
!        + B * ( T - T(LEFT) )
!        + C * ( T - T(LEFT) )**2
!        + D * ( T - T(LEFT) )**3
!
!    Here:
!      A = Y(LEFT)
!      B = ( Y(RIGHT) - Y(LEFT) ) / ( T(RIGHT) - T(LEFT) )
!        - ( YPP(RIGHT) + 2 * YPP(LEFT) ) * ( T(RIGHT) - T(LEFT) ) / 6
!      C = YPP(LEFT) / 2
!      D = ( YPP(RIGHT) - YPP(LEFT) ) / ( 6 * ( T(RIGHT) - T(LEFT) ) )
!
!  Modified:
!
!    24 February 2004
!
!  Author:
!
!    John Burkardt
!
!  Reference:
!
!    Carl de Boor,
!    A Practical Guide to Splines,
!    Springer Verlag, 1978.
!
!  Parameters:
!
!    Input, integer N, the number of knots.
!
!    Input, real ( kind = 8 ) T(N), the knot values.
!
!    Input, real ( kind = 8 ) Y(N), the data values at the knots.
!
!    Input, real ( kind = 8 ) YPP(N), the second derivatives of the spline at
!    the knots.
!
!    Input/output, integer LEFT, the suggested T interval to search.
!    LEFT should be between 1 and N-1.  If LEFT is not in this range,
!    then its value will be ignored.  On output, LEFT is set to the
!    actual interval in which TVAL lies.
!
!    Input, real ( kind = 8 ) TVAL, a point, typically between T(1) and T(N), at
!    which the spline is to be evalulated.  If TVAL lies outside
!    this range, extrapolation is used.
!
!    Output, real ( kind = 8 ) YVAL, YPVAL, YPPVAL, the value of the spline, and
!    its first two derivatives at TVAL.
!

!********************************************************************

  implicit none

  integer :: n, left, right

  real(8) dt, h, tval, yppval, ypval, yval

  real(8), dimension(n) :: t, y, ypp

!********************************************************************

!
!  Determine the interval [T(LEFT), T(RIGHT)] that contains TVAL.
!  
!  What you want from DVEC_BRACKET3 is that TVAL is to be computed
!  by the data in interval {T(LEFT), T(RIGHT)].
!
  call dvec_bracket3 ( n, t, tval, left )
  right = left + 1
!
!  In the interval LEFT, the polynomial is in terms of a normalized
!  coordinate  ( DT / H ) between 0 and 1.
!
  dt = tval - t(left)
  h = t(right) - t(left)

  yval = y(left) + dt * ( ( y(right) - y(left) ) / h &
              - ( ypp(right) / 6.0D+00 + ypp(left) / 3.0D+00 ) * h &
       + dt * ( 0.5D+00 * ypp(left) &
       + dt * ( ( ypp(right) - ypp(left) ) / ( 6.0D+00 * h ) ) ) )

  ypval = ( y(right) - y(left) ) / h &
      - ( ypp(right) / 6.0D+00 + ypp(left) / 3.0D+00 ) * h &
      + dt * ( ypp(left) &
      + dt * ( 0.5D+00 * ( ypp(right) - ypp(left) ) / h ) )

  yppval = ypp(left) + dt * ( ypp(right) - ypp(left) ) / h

  return
end subroutine

!--------------------------------------------------------------------

subroutine d3_np_fs ( n, a, b, x )

!
! D3_NP_FS factors and solves an D3 system.
!
!  Discussion:
!
!    The D3 storage format is used for a tridiagonal matrix.
!    The superdiagonal is stored in entries (1,2:N), the diagonal in
!    entries (2,1:N), and the subdiagonal in (3,1:N-1).  Thus, the
!    original matrix is "collapsed" vertically into the array.
!
!    This algorithm requires that each diagonal entry be nonzero.
!    It does not use pivoting, and so can fail on systems that
!    are actually nonsingular.
!
!  Example:
!
!    Here is how a D3 matrix of order 5 would be stored:
!
!       *  A12 A23 A34 A45
!      A11 A22 A33 A44 A55
!      A21 A32 A43 A54  *
!
!  Modified:
!
!    02 November 2003
!
!  Author:
!
!    John Burkardt
!
!  Parameters:
!
!    Input, integer N, the order of the linear system.
!
!    Input/output, real ( kind = 8 ) A(3,N).
!    On input, the tridiagonal matrix.
!    On output, the data in these vectors has been overwritten
!    by factorization information.
!
!    Input, real ( kind = 8 ) B(N), the right hand side of the linear system.
!
!    Output, real ( kind = 8 ) X(N), the solution of the linear system.
!

!********************************************************************

  implicit none

  integer :: n, i

  real (8), dimension(3,n) ::  a
  real (8), dimension(n) :: b, x
  real (8) :: xmult

!********************************************************************

!
!  The diagonal entries can't be zero.
!
  do i = 1, n
    if ( a(2,i) == 0.0D+00 ) then
      write ( *, '(a)' ) ' '
      write ( *, '(a)' ) 'D3_NP_FS - Fatal error!'
      write ( *, '(a,i6,a)' ) '  A(2,', i, ') = 0.'
      return
    end if
  end do

  x(1:n) = b(1:n)

  do i = 2, n
    xmult = a(3,i-1) / a(2,i-1)
    a(2,i) = a(2,i) - xmult * a(1,i)
    x(i)   = x(i)   - xmult * x(i-1)
  end do

  x(n) = x(n) / a(2,n)
  do i = n-1, 1, -1
    x(i) = ( x(i) - a(1,i+1) * x(i+1) ) / a(2,i)
  end do

  return
end subroutine

!--------------------------------------------------------------------

subroutine dvec_bracket ( n, x, xval, left, right )

!
! DVEC_BRACKET searches a sorted array for successive brackets of a value.
!
!  Discussion:
!
!    If the values in the vector are thought of as defining intervals
!    on the real line, then this routine searches for the interval
!    nearest to or containing the given value.
!
!  Modified:
!
!    06 April 1999
!
!  Author:
!
!    John Burkardt
!
!  Parameters:
!
!    Input, integer N, length of input array.
!
!    Input, real ( kind = 8 ) X(N), an array sorted into ascending order.
!
!    Input, real ( kind = 8 ) XVAL, a value to be bracketed.
!
!    Output, integer LEFT, RIGHT, the results of the search.
!    Either:
!      XVAL < X(1), when LEFT = 1, RIGHT = 2;
!      X(N) < XVAL, when LEFT = N-1, RIGHT = N;
!    or
!      X(LEFT) <= XVAL <= X(RIGHT).
!

!********************************************************************

  implicit none

  integer :: n, i, left, right

  real(8), dimension(n) :: x
  real(8) :: xval

!********************************************************************

  do i = 2, n - 1

    if ( xval < x(i) ) then
      left = i - 1
      right = i
      return
    end if

   end do

  left = n - 1
  right = n

  return
end subroutine

!--------------------------------------------------------------------

subroutine dvec_bracket3 ( n, t, tval, left )

!
! DVEC_BRACKET3 finds the interval containing or nearest a given value.
!
!  Discussion:
!
!    The routine always returns the index LEFT of the sorted array
!    T with the property that either
!    *  T is contained in the interval [ T(LEFT), T(LEFT+1) ], or
!    *  T < T(LEFT) = T(1), or
!    *  T > T(LEFT+1) = T(N).
!
!    The routine is useful for interpolation problems, where
!    the abscissa must be located within an interval of data
!    abscissas for interpolation, or the "nearest" interval
!    to the (extreme) abscissa must be found so that extrapolation
!    can be carried out.
!
!  Modified:
!
!    05 April 1999
!
!  Author:
!
!    John Burkardt
!
!  Parameters:
!
!    Input, integer N, length of the input array.
!
!    Input, real ( kind = 8 ) T(N), an array sorted into ascending order.
!
!    Input, real ( kind = 8 ) TVAL, a value to be bracketed by entries of T.
!
!    Input/output, integer LEFT.
!
!    On input, if 1 <= LEFT <= N-1, LEFT is taken as a suggestion for the
!    interval [ T(LEFT), T(LEFT+1) ] in which TVAL lies.  This interval
!    is searched first, followed by the appropriate interval to the left
!    or right.  After that, a binary search is used.
!
!    On output, LEFT is set so that the interval [ T(LEFT), T(LEFT+1) ]
!    is the closest to TVAL; it either contains TVAL, or else TVAL
!    lies outside the interval [ T(1), T(N) ].
!

!********************************************************************

  implicit none

  integer :: n, high, left, low, mid

  real(8), dimension(n) :: t
  real(8) :: tval

!********************************************************************

!
!  Check the input data.
!
  if ( n < 2 ) then
    write ( *, '(a)' ) ' '
    write ( *, '(a)' ) 'DVEC_BRACKET3 - Fatal error!'
    write ( *, '(a)' ) '  N must be at least 2.'
    stop
  end if
!
!  If LEFT is not between 1 and N-1, set it to the middle value.
!
  if ( left < 1 .or. n - 1 < left ) then
    left = ( n + 1 ) / 2
  end if
!
!  CASE 1: TVAL < T(LEFT):
!  Search for TVAL in [T(I), T(I+1)] for intervals I = 1 to LEFT-1.
!
  if ( tval < t(left) ) then

    if ( left == 1 ) then
      return
    else if ( left == 2 ) then
      left = 1
      return
    else if ( t(left-1) <= tval ) then
      left = left - 1
      return
    else if ( tval <= t(2) ) then
      left = 1
      return
    end if
!
!  ...Binary search for TVAL in [T(I), T(I+1)] for intervals I = 2 to LEFT-2.
!
    low = 2
    high = left - 2

    do

      if ( low == high ) then
        left = low
        return
      end if

      mid = ( low + high + 1 ) / 2

      if ( t(mid) <= tval ) then
        low = mid
      else
        high = mid - 1
      end if

    end do
!
!  CASE2: T(LEFT+1) < TVAL:
!  Search for TVAL in {T(I),T(I+1)] for intervals I = LEFT+1 to N-1.
!
  else if ( t(left+1) < tval ) then

    if ( left == n - 1 ) then
      return
    else if ( left == n - 2 ) then
      left = left + 1
      return
    else if ( tval <= t(left+2) ) then
      left = left + 1
      return
    else if ( t(n-1) <= tval ) then
      left = n - 1
      return
    end if
!
!  ...Binary search for TVAL in [T(I), T(I+1)] for intervals I = LEFT+2 to N-2.
!
    low = left + 2
    high = n - 2

    do

      if ( low == high ) then
        left = low
        return
      end if

      mid = ( low + high + 1 ) / 2

      if ( t(mid) <= tval ) then
        low = mid
      else
        high = mid - 1
      end if

    end do
!
!  CASE3: T(LEFT) <= TVAL <= T(LEFT+1):
!  T is in [T(LEFT), T(LEFT+1)], as the user said it might be.
!
  else

  end if

  return
end subroutine


double precision function xint(a, b, tol,rpar,ipar)

!
!  Modified
!	German Prieto
!	November 2004
!
!   Quadrature by Ierley's method of Chebychev sampling.
!
!   This is a slight variation of Gleen Ierly's code. What was
!   mainly done, was to avoid use of common blocks, defining all
!   variables and performing the numerical integration inside
!   (previously done by function pssevf.
!
!   Exponential convergence rate for analytic functions!  Much faster 
!   than Romberg; competitive with Gauss integration, without awkward 
!   weights.
! 
!   Integrates the function dpsw on (a, b) to absolute
!   accuracy tol > 0.
!
!   the function in time is given by rpar with ipar points
!  
!   I removed the optional printing routine part of the code, 
!   to make it easier to read. I also moved both nval, etol
!   as normal variables inside the routine.
!
!   nval = number of function calls made by routine
!   etol = approximate magnitude of the error of the result
! 
!   NB:  subroutine  set_xint  is called once before  xint  to
!     provide quadrature samples and weights.
!
!     I also altered the subroutine call, to get the weights
!     and not save them in a common block, but get them 
!     directly back.
!
!   lomx=number of samples = 2**nomx
!
!   calls set_xint
!

!********************************************************************

   implicit none

   integer, parameter :: nomx=8, lomx=256

   integer :: i, index, im, im2, n
   integer :: nval
   integer :: ipar   !, init

   real(8), dimension(ipar) :: rpar
   real(8) :: f1, f2, sq, etol, a, b, tol

   real(8), dimension(lomx+1) :: fv
!   real(8), dimension(nomx-1) :: est
   real(8), dimension(nomx)   :: est

!  set_xint subroutine variables

   real(8), dimension(nomx,lomx+1) :: w
   real(8), dimension(lomx+1)      :: x

!  function part

   real(8) :: om, ct, st, y

   real(8), parameter :: pi=3.141592653589793d0, tpi= 2.d0*pi

!********************************************************************


!  if (init.eq. 0) then 

   call set_xint(1,w,x)

!   endif

! 
!   Check tol
!

   if (tol .le. 0.0d0) then
      write(6, '(a)') 'In xint tol must be > 0'
      stop
   endif

   n=1
   im=2**(nomx+1)

   do index=1, nomx
      n=2*n
      im=im/2
      im2=im/2
      if (index.le.1) then
         do i=0, n
	    ! Bottom
	    y = a+(b-a)*x(im2*i+1)
	    om = tpi*y
	    call sft(rpar,ipar,om,ct,st)
	    f1 = ct*ct+st*st
	    ! Top
	    y = b-(b-a)*x(im2*i+1)
	    om = tpi*y
	    call sft(rpar,ipar,om,ct,st)
	    f2 = ct*ct+st*st
         
            fv(im2*i+1)= f1 + f2   
         enddo
      else
          do i=1, n-1, 2
	     ! Bottom
	     y = a+(b-a)*x(im2*i+1)
	     om = tpi*y
	     call sft(rpar,ipar,om,ct,st)
	     f1 = ct*ct+st*st
	     ! Top
	     y = b-(b-a)*x(im2*i+1)
	     om = tpi*y
	     call sft(rpar,ipar,om,ct,st)
	     f2 = ct*ct+st*st
 
             fv(im2*i+1)= f1 + f2
          enddo
      endif
      
      xint=0.0d0
      do i=0, n
         xint=xint+w(index, i+1)*fv(im2*i+1)
      enddo
      xint=xint*(b-a)
      est(index)=xint
      etol=0.0d0
 
! 
!   Check for convergence. 
!

      nval=2*n
 
      if (index .eq. 2) then
         if (est(index) .eq. est(index-1)) then
            return
         endif

      elseif (index > 2) then
         sq=(est(index)-est(index-1))**2
         etol=sq/(0.01d0*sq + abs(est(index)-est(index-2)))
         if (etol .le. tol) then
            return
         endif
      endif
      
   enddo

   write(6,*)' xint unable to provide requested accuracy'
 
   return
      
end function xint







subroutine set_xint(ising,w,x)

!
!  Modified
!	German Prieto
!	November 2004
!
!   Sets up weights and sample points for Ierley quadrature,
!
!   Slightly changed from original code, to avoid using common
!   blocks. Also avoided using some go to statements, not needed.
!
!   Input:
!	ising=1 	- integrand is analytic in closed interval
!	ising=2		- integrand may have bounded singularities
!			  at end points
!   
!   Returns: 
!   	w(nomx,lomx+1)	- weights
!    	x(lomx+1)	- sample points
!
!   lomx=number of samples = 2**nomx
!

!********************************************************************

   implicit none

   integer, parameter :: nomx=8, lomx=256
   real(8), parameter :: one=1.0d0, half=0.5d0
   
   integer :: k, i, nx, index, n, iprior, ising

   real(8) :: ck, rk, t, si, pin, pi

   real(8), dimension(nomx,lomx+1) :: w
   real(8), dimension(lomx+1)      :: x

!********************************************************************

   iprior = 0

   if (iprior .eq. ising) then 
      return
   endif
   
   iprior=ising

   pi=4.0d0*atan(one)
   n=2
   do index=1, nomx
      n=2*n
      nx=n-2
      if (index.eq.1) then
         nx=4
      endif
      pin=pi/dble(n)
      do i=0, n/2
         t=dble(i)*pin
         si=0.d0
      
         do k=0, nx, 2
            ck=4.0d0
            if (k.eq.0) then
  	       ck=2.0d0
  	    endif
            rk=dble(k)
            si=si+ck*cos(dble(k)*t)/(one-rk*rk)
         enddo

         if (i.eq.0.or.i.eq.n/2) then
	    si=half *si
	 endif

         t=cos(t)

	 if (ising == 2) then
            t=half *pi*(one +t)
            si=si*half * sin(t)*pi
            t=cos(t)
            x(i+1)=half *(one +t)
            w(index, i+1)=half *si/dble(n)
	 elseif (ising == 1) then
	    x(i+1)=half *(one +t)
            w(index, i+1)=half *si/dble(n)
         endif
      enddo
   enddo

   return
   
end subroutine set_xint










subroutine tridib ( n, eps1, d, e, e2, lb, ub, m11, m, w, ind, ierr )

!
!  Modified
!	German Prieto
!	November 2004
!
!  TRIDIB computes some eigenvalues of a real symmetric tridiagonal matrix.
!
!
!  Discussion:
!
!    This subroutine finds those eigenvalues of a tridiagonal
!    symmetric matrix between specified boundary indices,
!    using bisection.
!
!  Modified:
!
!    04 February 2003
!
!  Reference:
!
!    J H Wilkinson and C Reinsch,
!    Handbook for Automatic Computation,
!    Volume II, Linear Algebra, Part 2,
!    Springer Verlag, 1971.
!
!    B Smith, J Boyle, J Dongarra, B Garbow, Y Ikebe, V Klema, C Moler,
!    Matrix Eigensystem Routines, EISPACK Guide,
!    Lecture Notes in Computer Science, Volume 6,
!    Springer Verlag, 1976.
!
!  Parameters:
!
!    Input, integer N, the order of the matrix.
!
!    Input/output, real EPS1.  On input, an absolute error tolerance for 
!    the computed eigenvalues.  It should be chosen commensurate with
!    relative perturbations in the matrix elements of the order of the 
!    relative machine precision.  If the input EPS1 is non-positive, it 
!    is reset for each submatrix to a default value, namely, minus the
!    product of the relative machine precision and the 1-norm of the submatrix.
!
!    Input, real D(N), the diagonal elements of the input matrix.
!
!    Input, real E(N), the subdiagonal elements of the input matrix
!    in E(2:N).  E(1) is arbitrary.
!
!    Input/output, real E2(N).  On input, the squares of the corresponding 
!    elements of E.  E2(1) is arbitrary.  On output, elements of E2 
!    corresponding to elements of E regarded as negligible, have been 
!    replaced by zero, causing the matrix to split into a direct sum of 
!    submatrices.  E2(1) is also set to zero.
!
!    Input, integer M11, the lower boundary index for the desired eigenvalues.
!
!    Input, integer M, the number of eigenvalues desired.  The upper
!    boundary index M22 is then obtained as M22 = M11 + M - 1.
!
!    Output, real LB, UB, define an interval containing exactly the desired
!    eigenvalues.
!
!    Output, real W(M), the eigenvalues between indices M11 and M22 
!    in ascending order.
!
!    Output, integer IND(M), the submatrix indices associated with the 
!    corresponding eigenvalues in W: 1 for eigenvalues belonging to the 
!    first submatrix from the top, 2 for those belonging to the second 
!    submatrix, and so on.
!
!    Output, integer IERR, error flag.
!    0, for normal return,
!    3*N+1, if multiple eigenvalues at index M11 make unique selection 
!      impossible,
!    3*N+2, if multiple eigenvalues at index M22 make unique selection
!      impossible.
!

!********************************************************************

  implicit none

  integer, intent( in ) :: n
  integer :: m, i, ierr, ii, isturm, j, k, l
  integer :: m1, m11, m2, m22, p, q, r, s, tag 
  integer, dimension(m) :: ind
  
  real(8) :: eps1, lb, t1, t2, tst1, tst2
  real(8) :: u, ub, v, x0, x1, xu
  
  real(8), dimension(n) :: d, e, e2, rv4, rv5
  real(8), dimension(m) :: w 

!********************************************************************

  ierr = 0
  tag = 0
  xu = d(1)
  x0 = d(1)
  s = 0
  u = 0.0d0
!
!  Look for small sub-diagonal entries and determine an
!  interval containing all the eigenvalues.
!

  do i = 1, n

     x1 = u

     if ( i == n ) then
       u = 0.0d0
     else
       u = abs ( e(i+1) )
     end if

     xu = min ( xu, d(i)-(x1+u) )
     x0 = max ( x0, d(i)+(x1+u) )

     if ( i > 1 ) then
       tst1 = abs ( d(i) ) + abs ( d(i-1) )
       tst2 = tst1 + abs ( e(i) )
       if ( tst2 <= tst1 ) then
         e2(i) = 0.0d0
       end if
     else
       e2(i) = 0.0d0
     end if

  end do

  x1 = dble(n)
  x1 = x1 * max ( abs ( xu ), abs ( x0 ) ) * epsilon ( x1 )
  xu = xu - x1
  t1 = xu
  x0 = x0 + x1
  t2 = x0
!
!  Determine an interval containing exactly the desired eigenvalues.
!
  p = 1
  q = n
  m1 = m11 - 1
  if ( m1 == 0 ) go to 75
  isturm = 1

50 continue

  v = x1
  x1 = xu + (x0 - xu) * 0.5d0
  if ( x1 == v ) go to 980
  go to 320

60 continue

  if ( s - m1 ) 65, 73, 70

65 continue

  xu = x1
  go to 50

70 continue

  x0 = x1
  go to 50

73 continue

  xu = x1
  t1 = x1

75 continue

  m22 = m1 + m
  if ( m22 == n ) go to 90
  x0 = t2
  isturm = 2
  go to 50

80 continue

   if ( s - m22 ) 65, 85, 70

85 continue

   t2 = x1

90 continue

  q = 0
  r = 0
!
!  Establish and process next submatrix, refining interval by the 
!  Gerschgorin bounds.
!
100 continue

  if ( r == m ) then
    go to 1001
  end if

  tag = tag + 1
  p = q + 1
  xu = d(p)
  x0 = d(p)
  u = 0.0d0

  do q = p, n

    x1 = u
    u = 0.0d0
    v = 0.0d0

    if ( q < n ) then
      u = abs ( e(q+1) )
      v = e2(q+1)
    end if

    xu = min ( d(q)-(x1+u), xu )
    x0 = max ( d(q)+(x1+u), x0 )

    if ( v == 0.0d0 ) then
      exit
    end if

  end do

  x1 = max ( abs ( xu ), abs ( x0 ) ) * epsilon ( x1 )
  if ( eps1 <= 0.0d0 ) eps1 = -x1
  if ( p /= q ) go to 180
!
!  Check for isolated root within interval.
!
  if ( t1 > d(p) .or. d(p) >= t2 ) go to 940
  m1 = p
  m2 = p
  rv5(p) = d(p)
  go to 900

180 continue

  x1 = x1 * (q - p + 1)
  lb = max ( t1, xu-x1 )
  ub = min ( t2, x0+x1 )
  x1 = lb
  isturm = 3
  go to 320

200 continue

  m1 = s + 1
  x1 = ub
  isturm = 4
  go to 320

220 continue

  m2 = s
  if ( m1 > m2 ) go to 940
!
!  Find roots by bisection.
!
  x0 = ub
  isturm = 5

  rv5(m1:m2) = ub
  rv4(m1:m2) = lb
!
!  Loop for the K-th eigenvalue.
!
  k = m2

250 continue

  xu = lb

  do ii = m1, k

    i = m1 + k - ii
    if ( xu < rv4(i) ) then
      xu = rv4(i)
      exit
    end if

  end do

  if ( x0 > rv5(k) ) x0 = rv5(k)
!
!  Next bisection step.
!
300  continue

     x1 = ( xu + x0 ) * 0.5d0
     if ( ( x0 - xu ) <= abs ( eps1) ) go to 420
     tst1 = 2.0d0 * ( abs ( xu ) + abs ( x0 ) )
     tst2 = tst1 + (x0 - xu)
     if ( tst2 == tst1 ) go to 420
!
!  Sturm sequence.
!
320  continue

     s = p - 1
     u = 1.0d0

     do i = p, q

       if ( u == 0.0d0 ) then
         v = abs ( e(i) ) / epsilon ( v )
         if ( e2(i) == 0.0d0 ) v = 0.0d0
       else
         v = e2(i) / u
       end if

       u = d(i) - x1 - v

       if ( u < 0.0d0 ) then
         s = s + 1
       end if

     end do

     go to (60,80,200,220,360), isturm
!
!  Refine intervals.
!
360  continue

     if ( s >= k) go to 400
     xu = x1
     if ( s >= m1) go to 380
     rv4(m1) = x1
     go to 300

380  continue

     rv4(s+1) = x1
     if ( rv5(s) > x1) rv5(s) = x1
     go to 300

400  continue

     x0 = x1
     go to 300
!
!  K-th eigenvalue found.
!
420  continue

  rv5(k) = x1
  k = k - 1
  if ( k >= m1 ) go to 250
!
!  Order eigenvalues tagged with their submatrix associations.
!
900 continue

  s = r
  r = r + m2 - m1 + 1
  j = 1
  k = m1

  do l = 1, r

     if ( j > s ) go to 910
     if ( k > m2 ) go to 940
     if ( rv5(k) >= w(l) ) go to 915

     do ii = j, s
       i = l + s - ii
       w(i+1) = w(i)
       ind(i+1) = ind(i)
     end do

910  continue

     w(l) = rv5(k)
     ind(l) = tag
     k = k + 1
     go to 920

915  continue

     j = j + 1

920  continue

  end do

940 continue

  if ( q < n ) go to 100

  go to 1001
!
!  Set error: interval cannot be found containing exactly the 
!  desired eigenvalues.
!
980 continue

  ierr = 3 * n + isturm

1001 continue

  lb = t1
  ub = t2
  return

end subroutine tridib






subroutine tinvit ( n, d, e, e2, m, w, ind, z, ierr )

!
!  Modified
!	German Prieto
!	November 2004
!
!  TINVIT computes eigenvectors from eigenvalues, real tridiagonal symmetric.
!
!
!  Discussion:
!
!    This subroutine finds those eigenvectors of a tridiagonal
!    symmetric matrix corresponding to specified eigenvalues,
!    using inverse iteration.
!
!  Modified:
!
!    04 February 2003
!
!  Reference:
!
!    J H Wilkinson and C Reinsch,
!    Handbook for Automatic Computation,
!    Volume II, Linear Algebra, Part 2,
!    Springer Verlag, 1971.
!
!    B Smith, J Boyle, J Dongarra, B Garbow, Y Ikebe, V Klema, C Moler,
!    Matrix Eigensystem Routines, EISPACK Guide,
!    Lecture Notes in Computer Science, Volume 6,
!    Springer Verlag, 1976.
!
!  Parameters:
!
!    Input, integer N, the order of the matrix.
!
!    Input, real D(N), the diagonal elements of the matrix.
!
!    Input, real E(N), contains the subdiagonal elements of the input matrix
!    in E(2:N).  E(1) is arbitrary.
!
!    Input, real E2(N), contains the squares of the corresponding elements 
!    of E, with zeros corresponding to negligible elements of E.
!    E(I) is considered negligible if it is not larger than the product of 
!    the relative machine precision and the sum of the magnitudes of D(I) 
!    and D(I-1).  E2(1) must contain 0.0E+00 if the eigenvalues are in 
!    ascending order, or 2.0E+00 if the eigenvalues are in descending order.  
!    If BISECT, TRIDIB, or IMTQLV has been used to find the eigenvalues,
!    their output E2 array is exactly what is expected here.
!
!    Input, integer M, the number of specified eigenvalues.
!
!    Input, real W(M), the eigenvalues.
!
!    Input, integer IND(M), the submatrix indices associated with the 
!    corresponding eigenvalues in W: 1 for eigenvalues belonging to the 
!    first submatrix from the top, 2 for those belonging to the second 
!    submatrix, and so on.
!
!    Output, real Z(N,M), the associated set of orthonormal eigenvectors.
!    Any vector which fails to converge is set to zero.
!
!    Output, integer IERR, error flag.
!    0, for normal return,
!    -R, if the eigenvector corresponding to the R-th eigenvalue fails to 
!      converge in 5 iterations.
!

!********************************************************************

   implicit none

   integer :: m,n, group, i, ierr, ii, ip, its, j, jj
   integer :: p, q, r, s, tag 

   integer, dimension(m) :: ind
 
   real(8) :: eps2, eps3, eps4, norm, order
   real(8) :: u, uk, v, x0, x1, xu
   real(8), dimension(n) :: d, e, e2, rv1, rv2, rv3, rv4, rv6
   real(8), dimension(m) :: w
   real(8), dimension(n,m) :: z

!********************************************************************

  ierr = 0

  if ( m == 0 ) then
    return
  end if

  u = 0.0d0
  x0 = 0.0d0

  tag = 0
  order = 1.0d0 - e2(1)
  q = 0
!
!  Establish and process next submatrix.
!
100 continue

  p = q + 1

  do q = p, n
    if ( q == n ) then
      exit
    end if
    if ( e2(q+1) == 0.0d0 ) then
      exit
    end if
  end do
!
!  Find vectors by inverse iteration.
!
140 continue

  tag = tag + 1
  s = 0

  do r = 1, m

     if ( ind(r) /= tag ) go to 920

     its = 1
     x1 = w(r)

     if ( s /= 0 ) go to 510
!
!  Check for isolated root.
!
     xu = 1.0d0

     if ( p == q ) then
       rv6(p) = 1.0d0
       go to 870
     end if

     norm = abs ( d(p) )
     ip = p + 1

     do i = p+1, q
       norm = max ( norm, abs ( d(i) ) + abs ( e(i) ) )
     end do
!
!  EPS2 is the criterion for grouping,
!  EPS3 replaces zero pivots and equal roots are modified by EPS3,
!  EPS4 is taken very small to avoid overflow.
!
     eps2 = 0.001d0 * norm
     eps3 = abs ( norm ) * epsilon ( eps3 )
     uk = dble(q - p + 1)
     eps4 = uk * eps3
     uk = eps4 / sqrt ( uk )
     s = p

505 continue

     group = 0
     go to 520
!
!  Look for close or coincident roots.
!
510  continue

     if ( abs ( x1 - x0 ) >= eps2 ) go to 505

     group = group + 1

     if ( order * (x1 - x0) <= 0.0d0 ) then
       x1 = x0 + order * eps3
     end if
!
!  Elimination with interchanges and initialization of vector.
!
520  continue

     v = 0.0d0

     do i = p, q

        rv6(i) = uk

        if ( i == p ) go to 560

        if ( abs ( e(i) ) < abs ( u ) ) go to 540

        xu = u / e(i)
        rv4(i) = xu
        rv1(i-1) = e(i)
        rv2(i-1) = d(i) - x1
        rv3(i-1) = 0.0d0
        if ( i /= q ) rv3(i-1) = e(i+1)
        u = v - xu * rv2(i-1)
        v = - xu * rv3(i-1)
        go to 580

540     continue

        xu = e(i) / u
        rv4(i) = xu
        rv1(i-1) = u
        rv2(i-1) = v
        rv3(i-1) = 0.0d0

560     continue

        u = d(i) - x1 - xu * v
        if ( i /= q ) v = e(i+1)

580     continue

     end do

     if ( u == 0.0d0 ) then
       u = eps3
     end if

     rv1(q) = u
     rv2(q) = 0.0d0
     rv3(q) = 0.0d0
!
!  Back substitution.
!
600   continue

  do ii = p, q
    i = p + q - ii
    rv6(i) = ( rv6(i) - u * rv2(i) - v * rv3(i) ) / rv1(i)
    v = u
    u = rv6(i)
  end do
!
!  Orthogonalize with respect to previous members of group.
!
     j = r

     do jj = 1, group

       do

         j = j - 1

         if ( ind(j) == tag ) then
           exit
         end if

       end do

       xu = dot_product ( rv6(p:q), z(p:q,j) )

       rv6(p:q) = rv6(p:q) - xu * z(p:q,j)

     end do

     norm = sum ( abs ( rv6(p:q) ) )

     if ( norm >= 1.0d0 ) go to 840
!
!  Forward substitution.
!
     if ( its == 5 ) go to 830

     if ( norm == 0.0d0 ) then
       rv6(s) = eps4
       s = s + 1
       if ( s > q ) s = p
       go to 780
     end if

740  continue

     xu = eps4 / norm
     rv6(p:q) = rv6(p:q) * xu
!
!  Elimination operations on next vector iterate.
!
780  continue
!
!  If RV1(I-1) == E(I), a row interchange was performed earlier in the
!  triangularization process.
!
     do i = ip, q

       u = rv6(i)

       if ( rv1(i-1) == e(i) ) then
         u = rv6(i-1)
         rv6(i-1) = rv6(i)
       end if

       rv6(i) = u - rv4(i) * rv6(i-1)

     end do

     its = its + 1
     go to 600
!
!  Set error: non-converged eigenvector.
!
830  continue

     ierr = -r
     xu = 0.0d0
     go to 870
!
!  Normalize so that sum of squares is 1 and expand to full order.
!
840  continue

     u = 0.0d0
     do i = p, q
       u = pythag ( u, rv6(i) )
     end do

     xu = 1.0d0 / u

870  continue

     z(1:n,r) = 0.0d0
     z(p:q,r) = rv6(p:q) * xu

     x0 = x1

920  continue

  end do

  if ( q < n ) go to 100

  return

end subroutine tinvit









double precision function pythag ( a, b )

!
!  Modified
!	German Prieto
!	November 2004
!
!  Modified
!	German Prieto
!	March 2006
!	Added rpythag, the real(4) version for QI theory to use
!
!
!  PYTHAG computes SQRT ( A**2 + B**2 ) carefully.
!
!
!  Discussion:
!
!    The formula
!
!      PYTHAG = sqrt ( A**2 + B**2 )
!
!    is reasonably accurate, but can fail if, for example, A**2 is larger 
!    than the machine overflow.  The formula can lose most of its accuracy 
!    if the sum of the squares is very large or very small.
!
!  Modified:
!
!    04 February 2003
!
!  Reference:
!
!    J H Wilkinson and C Reinsch,
!    Handbook for Automatic Computation,
!    Volume II, Linear Algebra, Part 2,
!    Springer Verlag, 1971.
!
!    B Smith, J Boyle, J Dongarra, B Garbow, Y Ikebe, V Klema, C Moler,
!    Matrix Eigensystem Routines, EISPACK Guide,
!    Lecture Notes in Computer Science, Volume 6,
!    Springer Verlag, 1976.
!
!  Modified:
!
!    04 February 2003
!
!  Parameters:
!
!    Input:
!    	a   
!    	b  	- the two legs of a right triangle.
!
!    Output:
!    
!    	pythag	- the length of the hypotenuse.
!

!********************************************************************

   implicit none

   real(8) :: a, b, p, r, s, t, u

!********************************************************************

  p = max ( abs ( a ), abs ( b ) )

  if ( p /= 0.0d0 ) then

    r = ( min ( abs ( a ), abs ( b ) ) / p )**2

    do

      t = 4.0d0 + r

      if ( t == 4.0d0 ) then
        exit
      end if

      s = r / t
      u = 1.0d0 + 2.0d0 * s
      p = u * p
      r = ( s / u )**2 * r

    end do

  end if

  pythag = p

  return

end function pythag

!--------------------------------------------------------------------

real function rpythag(a,b)

!********************************************************************

   implicit none

   real(4) :: a, b
   real(4) :: p, r, s, t, u

!********************************************************************

  p = max ( abs ( a ), abs ( b ) )

  if ( p /= 0.0d0 ) then

    r = ( min ( abs ( a ), abs ( b ) ) / p )**2

    do

      t = 4.0 + r

      if ( t == 4.0 ) then
        exit
      end if

      s = r / t
      u = 1.0 + 2.0 * s
      p = u * p
      r = ( s / u )**2 * r

    end do

  end if

  rpythag = p

  return

end function rpythag

!--------------------------------------------------------------------




subroutine sft(x,n,om,ct,st)

!
!  Modified
!	German Prieto
!	November 2004
!
!   calculates fourier transform of real sequence x(i),i=1,...n
!   at angular frequency om normalized so that nyquist=pi. the sine
!   transform is returned in st and the cosine transform in ct.
!   algorithm is that of goertzal with modifications by
!   gentleman, comp.j. 1969
!   transform is not normalized
!   to normalize one-sided ft, divide by sqrt(data length)
!   for positive om, the ft is defined as ct-(0.,1.)st or like slatec
!   cfftf
! 

!**********************************************************************

   implicit none

   real(8), parameter :: pi=3.141592653589793238d0, tp=2.d0*pi

   integer :: k, n, np1, l

   real(8) :: b, e, d, c, a, s, om, ct, st 
 
   real(8), dimension(n) :: x

!**********************************************************************

   np1=n+1
   l=6.d0*om/tp
   s=sin(om)
   a=0.d0
   c=0.d0
   d=0.d0
   e=0.d0
   
   if (l.eq.0)then

!     recursion for low frequencies (.lt. nyq/3)

      b=-4.d0*sin(om/2.d0)**2
      do k=1,n
         c=a
         d=e
         a=x(np1-k)+b*d+c
         e=a+d
      enddo
      
   elseif (l.eq.1)then

!     regular goertzal algorithm for intermediate frequencies

      b=2.d0*cos(om)
      do  k=1,n
          a=x(np1-k)+b*e-d
          d=e
          e=a
      enddo
      
   else

!     recursion for high frequencies (.gt. 2*nyq/3)
      
      b=4.d0*cos(om/2.d0)**2
        do k=1,n
           c=a
           d=e
           a=x(np1-k)+b*d-c
           e=a-d
	enddo
   
   endif
   
   st=-s*d
   ct=a-b*d/2.d0
      
   return
      
end subroutine sft






end module
