module tapers_mod
  ! taper types: 
  ! 1) 'zero'            --  needed: 'nothing' 
  ! 2) 'rect'            --  needed: t1,t2 
  ! 3) 'linear_up'       --  needed: t1,t2
  ! 4) 'linear_down'     --  needed: t1,t2
  ! 5) 'pyramid'         --  needed: t1,t2
  ! 6) 'pyramid_band'    --  needed: t1,t2,t3,t4
  ! 7) 'hann'            --  needed: t1,t2
  ! 8) 'hann_half'       --  needed: t1,t2
  ! 9) 'hann_band'       --  needed: t2,t2,t3,t4
  ! 10) 'cosine'         --  needed: t1,t2
  ! 11) 'gaussian'       --  needed: t1,t2,sigma
  ! 12) 'gaussian_half'  --  needed: t1,t2,sigma
  ! 13) 'gaussian_block' --  needed: t1,t2,t3,t4,sigma



  ! Contains a variety of callable tapers
  ! initialize by n1, nt, t1, t2, inv
  ! n1: length of taper
  ! nt: number of tapers ( for slepian only at this point )
  ! t1: index of tie point one    ( effective shape length : t2-(t1-1) )
  ! t2: index of tie point two    ( taper(t1:t2) = taper )
  ! inv: taper = 1.-taper         ( 'invert' the taper shape )
  ! This module basically, initializes the taper_mod module,
  ! Multiple calls can construct a more complicated taper.
  !
  ! Very usefull: http://en.wikipedia.org/wiki/Window_function
  ! Last accessed: 2/15/2010

  !The following tapers have non zero boundary points:

  ! Rect-likes
  ! Gaussian-likes
  ! 

  ! The following tapers have periodic boundary points (at t1 : 0 at t2 ~: 0):
  ! cosine-like
  ! half_hann-like
  ! Linear-likes (but linear_down at t1 ~: 0 at t2 : 0)
  ! hann-like & Pyramid-likes (at t1 : 0 at t2 : 0):
  ! gaussian (at t1 ~: 0 at t2 ~: 0)


  use addnull_mod
  use   tape_mod
  use mdtape_mod
  
  implicit none
  
  ! For internal itterated use to fill taper1,2,3
  integer, private              :: n
  real, private,allocatable     :: taper(:),taper_com(:)

  ! For internal use to store taper1,2,3
  integer, private              :: n1,n2,n3
  real, private,allocatable     :: taper1(:),taper2(:),taper3(:)

  real, parameter, private      :: pi=3.14159265


  ! Interface, to get the taper(s) collected
  interface tapers_gettaper
    module procedure tapers_gettaper_1d
    module procedure tapers_gettaper_2d
    module procedure tapers_gettaper_3d
  end interface 

  ! Interface, to get the taper(s) collected
  interface tapers_invert
    module procedure tapers_invert1
    module procedure tapers_invert2
    module procedure tapers_invert3
  end interface 

  contains
  
  ! (Re)Initialize this module
  subroutine tapers_init(n1_in,n2_in,n3_in)
    integer, intent(in),optional  :: n1_in,n2_in,n3_in

    if (present( n1_in)) then; n1  = n1_in;  else; n1 = 1;  end if
    if (present( n2_in)) then; n2  = n2_in;  else; n2 = 1;  end if
    if (present( n3_in)) then; n3  = n3_in;  else; n3 = 1;  end if

    if (.not. allocated(taper1)) then; allocate( taper1(n1) ); taper1  = 1.; end if
    if (.not. allocated(taper2)) then; allocate( taper2(n2) ); taper2  = 1.; end if
    if (.not. allocated(taper3)) then; allocate( taper3(n3) ); taper3  = 1.; end if
    
  end subroutine

  subroutine tapers_new_init(n_in)
    integer, intent(in),optional  :: n_in
    
    if (present( n_in))  then; n   = n_in;   else; n  = 1;  end if

    if (allocated( taper     )) then; deallocate( taper    ); end if;
    if (allocated( taper_com )) then; deallocate( taper_com); end if;

    allocate( taper    (n) ); call addnull(.false.,taper    );
    allocate( taper_com(n) ); call addnull(.false.,taper_com);
    
  end subroutine


  ! Close this module
  subroutine tapers_close()
    if (allocated( taper     )) then; deallocate( taper    ); end if;
    if (allocated( taper_com )) then; deallocate( taper_com); end if;
    if (allocated( taper1    )) then; deallocate( taper1   ); end if;
    if (allocated( taper2    )) then; deallocate( taper2   ); end if;
    if (allocated( taper3    )) then; deallocate( taper3   ); end if;
  end subroutine
  

 ! ! To just getting a 1d taper, for external use
 ! subroutine tapers_gettaper(taper_out,inv_in,dir_in)
 !   real,intent(out)            :: taper_out(:)
 !   logical,intent(in),optional :: inv_in
 !   integer,intent(in),optional :: dir_in
 !   logical                     :: inv
 !   integer                     :: dir

 !   if (present(inv_in)) then; inv  = inv_in;  else; inv=.false.; end if
 !   if (present(dir_in)) then; dir  = dir_in;  else; dir=1      ; end if
    
 !   if (dir==1) then; taper_out = taper1; end if
 !   if (dir==2) then; taper_out = taper2; end if
 !   if (dir==3) then; taper_out = taper3; end if
 !   call tapers_invert(inv,taper_out)
    
 ! end subroutine


  ! Compile a 1d md taper from 1d tapers
  subroutine tapers_gettaper_1d(taper_out,inv_in)
    real,intent(out)            :: taper_out(:)
    logical,intent(in),optional :: inv_in
    logical                     :: inv
    real,allocatable            :: tapermd(:)
    integer                     :: ii

    if (present(inv_in)) then; inv  = inv_in;  else; inv=.false.; end if

    allocate (tapermd(n1))

    do ii=1,n1;
      tapermd(ii) = taper1(ii)
    end do

    call tapers_invert(inv,tapermd)
    taper_out = tapermd

    deallocate(tapermd)
    
  end subroutine

  ! Compile a 2d md taper from 1d tapers
  subroutine tapers_gettaper_2d(taper_out,inv_in)
    real,intent(out)            :: taper_out(:,:)
    logical,intent(in),optional :: inv_in
    logical                     :: inv
    real,allocatable            :: tapermd(:,:)
    integer                     :: ii,jj

    if (present(inv_in)) then; inv  = inv_in;  else; inv=.false.; end if

    allocate (tapermd(n1,n2))
    do jj=1,n2;
    do ii=1,n1;
      tapermd(ii,jj) = taper1(ii)*taper2(jj)
    end do
    end do

    call tapers_invert(inv,tapermd)
    taper_out = tapermd

    deallocate(tapermd)
    
  end subroutine

  ! To just getting a 1d taper, for external use
  subroutine tapers_gettaper_3d(taper_out,inv_in)
    real,intent(out)            :: taper_out(:,:,:)
    logical,intent(in),optional :: inv_in
    logical                     :: inv
    real,allocatable            :: tapermd(:,:,:)
    integer                     :: ii,jj,kk

    if (present(inv_in)) then; inv  = inv_in;  else; inv=.false.; end if

    allocate (tapermd(n1,n2,n3))
    do kk=1,n3;
    do jj=1,n2;
    do ii=1,n1;
      tapermd(ii,jj,kk) = taper1(ii)*taper2(jj)*taper3(kk)
    end do
    end do
    end do

    call tapers_invert(inv,tapermd)
    taper_out = tapermd

    deallocate(tapermd)
    
  end subroutine
  
  
  
  ! Subroutine to pass constructed taper into taper1,2,3
  subroutine  tapers_insert(add,t1_in,t2_in,dir_in,inv_in)
    logical, intent(in)            :: add
    integer,optional,intent(in)    :: t1_in,t2_in,dir_in
    logical, intent(in),optional   :: inv_in
    integer                        :: t1,t2,dir,ii
    logical                        :: inv

    if (present(t1_in )) then; t1  = t1_in;  else; t1 =1;       end if
    if (present(t2_in )) then; t2  = t2_in;  else; t2 =n;       end if
    if (present(dir_in)) then; dir = dir_in; else; dir=1;       end if
    if (present(inv_in)) then; inv = inv_in; else; inv=.false.; end if

    call tapers_invert(inv,taper_com)
    if (dir==1) then
      if (.not. size(taper_com) == n1) then; write(0,*) "size(taper_in)!=size(taper) in next call to taper_mod"; call exit(); end if      
      call addnull(add,taper1(t1:t2))
      do ii=t1,t2;
        taper1(ii) = taper1(ii) + taper_com(ii)
      end do
    end if
    if (dir==2) then
      if (.not. size(taper_com) == n2) then; write(0,*) "size(taper_in)!=size(taper) in next call to taper_mod"; call exit(); end if      
      call addnull(add,taper2(t1:t2))
      do ii=t1,t2;
        taper2(ii) = taper2(ii) + taper_com(ii)
      end do
    end if
    if (dir==3) then
      if (.not. size(taper_com) == n3) then; write(0,*) "size(taper_in)!=size(taper) in next call to taper_mod"; call exit(); end if      
      call addnull(add,taper3(t1:t2))
      do ii=t1,t2;
        taper3(ii) = taper3(ii) + taper_com(ii)
      end do
    end if
    call tapers_invert(inv,taper_com)
    
  end subroutine
  
  
  subroutine tapers_insert_tape(add_in,inv_in,t1_in,t2_in,dir_in)
    logical,optional,intent(in)    :: inv_in,add_in
    integer,optional,intent(in)    :: t1_in,t2_in,dir_in
    integer                        :: t1,t2,dir
    logical                        :: inv,add
    
    if (present(inv_in)) then; inv = inv_in; else; inv=.false.; end if
    if (present(add_in)) then; add = add_in; else; add=.false.; end if
    if (present(dir_in)) then; dir = dir_in; else; dir=0; end if;
    
    if (dir==0) then
      call tapers_invert (inv,taper1)
      call tapers_invert (inv,taper2)
      call tapers_invert (inv,taper3)
      call tape_insert   (add,taper1,1,n1,dir_in=1)
      call tape_insert   (add,taper2,1,n2,dir_in=2)
      call tape_insert   (add,taper3,1,n3,dir_in=3)
      call tapers_invert (inv,taper1)
      call tapers_invert (inv,taper2)
      call tapers_invert (inv,taper3)
    end if

    if (dir==1) then
      if (present(t1_in )) then; t1  = t1_in;  else; t1 =1; end if
      if (present(t2_in )) then; t2  = t2_in;  else; t2 =n1; end if
      call tapers_invert (inv,taper1)
      call tape_insert   (add,taper1,t1,t2,dir_in=1)
      call tapers_invert (inv,taper1)
    end if
    if (dir==2) then
      if (present(t1_in )) then; t1  = t1_in;  else; t1 =1; end if
      if (present(t2_in )) then; t2  = t2_in;  else; t2 =n2; end if
      call tapers_invert (inv,taper2)
      call tape_insert   (add,taper2,t1,t2,dir_in=2)
      call tapers_invert (inv,taper2)
    end if
    if (dir==3) then
      if (present(t1_in )) then; t1  = t1_in;  else; t1 =1; end if
      if (present(t2_in )) then; t2  = t2_in;  else; t2 =n3; end if
      call tapers_invert (inv,taper3)
      call tape_insert   (add,taper3,t1,t2,dir_in=3)
      call tapers_invert (inv,taper3)
    end if

  end subroutine


  subroutine tapers_insert_mdtape(add_in,inv_in,t1_in,t2_in,dir_in)
    logical,optional,intent(in)    :: inv_in,add_in
    integer,optional,intent(in)    :: t1_in(3),t2_in(3),dir_in
    integer                        :: t1(3),t2(3),dir
    logical                        :: inv,add
    

    real,allocatable            :: tapermd(:,:,:)
    integer                     :: ii,jj,kk

    if (present(inv_in)) then; inv = inv_in; else; inv=.false.; end if
    if (present(add_in)) then; add = add_in; else; add=.false.; end if
    if (present(t1_in )) then;  t1 = t1_in;  else;  t1=(/1, 1, 1 /); end if
    if (present(t2_in )) then;  t2 = t2_in;  else;  t2=(/n1, n2, n3 /); end if

    allocate (tapermd(n1,n2,n3))
    do kk=1,n3;
    do jj=1,n2;
    do ii=1,n1;
      tapermd(ii,jj,kk) = taper1(ii)*taper2(jj)*taper3(kk)
    end do;
    end do;
    end do;

    call tapers_invert (inv,tapermd)
    call mdtape_insert (add,tapermd,t1,t2)

    deallocate(tapermd)


  end subroutine


  subroutine tapers_interface(ntapers, dir_in, t1_in,t2_in,t3_in,t4_in,type_in,sigma_in,scale_in,inv_in)
    integer                       :: ntapers
    integer, intent(in), optional :: t1_in(ntapers),t2_in(ntapers),t3_in(ntapers),t4_in(ntapers),type_in(ntapers),dir_in(ntapers)
    real,    intent(in), optional :: sigma_in(ntapers), scale_in(ntapers)
    logical, intent(in), optional :: inv_in(ntapers)

    integer                       :: t1(ntapers),t2(ntapers),t3(ntapers),t4(ntapers), type(ntapers),dir(ntapers)
    real                          :: sigma(ntapers), scale(ntapers)
    logical                       :: inv(ntapers)

    integer                       :: ii,n(ntapers)


    if (present(t1_in)) then; t1 = t1_in; else;
      do ii=1,ntapers;   dir(ii) = 1; end do; end if;
    if (present(t1_in)) then; t1 = t1_in; else;
      do ii=1,ntapers;        t1 = 1; end do; end if;
    if (present(t2_in)) then; t2 = t2_in; else;
      do ii=1,ntapers;        t2 = n1; end do; end if;
    if (present(t3_in)) then; t3 = t3_in; else;
      do ii=1,ntapers;        t3 = n1; end do; end if;
    if (present(t4_in)) then; t4 = t4_in; else;
      do ii=1,ntapers;        t4 = n1; end do; end if;
    if (present( type_in)) then;  type =  type_in; else;
      do ii=1,ntapers;  type(ii) = 0; end do; end if;
    if (present(sigma_in)) then; sigma = sigma_in; else;
      do ii=1,ntapers; sigma(ii) = 0; end do; end if;
    if (present(  inv_in)) then;   inv =   inv_in; else;
      do ii=1,ntapers;   inv(ii) = .false.; end do; end if;
    if (present(scale_in)) then; scale = scale_in; else;
      do ii=1,ntapers; scale(ii) = .false.; end do; end if;
    if (present(  dir_in)) then;   dir = dir_in; else;
      do ii=1,ntapers;   dir(ii) = 1;       end do; end if;

    do ii=1,ntapers;
      if (dir(ii)==1) then; n(ii)=n1; end if
      if (dir(ii)==2) then; n(ii)=n2; end if
      if (dir(ii)==3) then; n(ii)=n3; end if
    end do

    do ii=1,ntapers
      if (t1(ii) .gt. t2(ii)) then; write(0,*) "Error - t2 < t1",ii,"-taper"; call exit(); end if
      if (t2(ii) .gt. t3(ii)) then; write(0,*) "Error - t3 < t2",ii,"-taper"; call exit(); end if
      if (t3(ii) .gt. t4(ii)) then; write(0,*) "Error - t3 < t2",ii,"-taper"; call exit(); end if
      call tapers_new_init(n_in=n(ii))
      call tapers_create  (t1_in=t1(ii),t2_in=t2(ii),t3_in=t3(ii),t4_in=t4(ii),type_in=type(ii),sigma_in=sigma(ii),scale_in=scale(ii))
      call tapers_insert  (.false.,1,n(ii),dir_in=dir(ii),inv_in=inv_in(ii))
    end do

  end subroutine




  subroutine tapers_create(t1_in,t2_in,t3_in,t4_in,type_in,sigma_in,scale_in)

    real   , intent(in),optional :: sigma_in
    integer, intent(in),optional :: t1_in,t2_in,t3_in,t4_in,type_in
    real                         :: sigma
    integer                      :: t1,t2,t3,t4
    integer                      :: type
    real, optional,intent(in)    :: scale_in
    real                         :: scale

    if (present( t1_in)) then;  t1 = t1_in ; else;  t1 = 1; end if;
    if (present( t2_in)) then;  t2 = t2_in ; else;  t2 = n; end if;
    if (present( t3_in)) then;  t3 = t3_in ; else;  t3 = n; end if;
    if (present( t4_in)) then;  t4 = t4_in ; else;  t4 = n; end if;
    if (present(type_in )) then; type  = type_in;  else; type  = 1;       end if;
    if (present(sigma_in)) then; sigma = sigma_in; else; sigma = .4;      end if;
    if (present(scale_in)) then; scale = scale_in; else; scale=1.; end if

  select case(type)
    
    case(1)
      call tapers_zero(                    )
    case(2)
      call tapers_rect  (.false.,t1  , t2, scale_in=scale)
    case(3)
      call tapers_linear_up(.false.,t1  , t2, scale_in=scale)
    case(4)
      call tapers_linear_down(.false.,t1  , t2, scale_in=scale)
    case(5)
      call tapers_pyramid(.false.,t1  , t2, scale_in=scale)
    case(6)
      call tapers_pyramid_block(.false., t1,t2,t3,t4, scale_in=scale)
    case(7)
      call tapers_hann(.false.,t1, t2, scale_in=scale)
    case(8)
      call tapers_hann_half( .false.,t1, t2, scale_in=scale)
    case(9)
      call tapers_hann_block(.false., t1,t2,t3,t4, scale_in=scale)  ! repair, using a half_hann subroutine
    case(10)
      call tapers_cosine   (.false.,t1, t2, scale_in=scale)
    case(11)
      call tapers_gaussian (.false.,t1, t2, scale_in=scale, sigma_in=sigma )
    case(12)
      call tapers_gaussian_half(.false.,t1, t2, scale_in=scale, sigma_in=sigma )
    case(13)
      call tapers_gaussian_block(.false.,t1, t2, t3, t4, scale_in=scale, sigma_in=sigma )! repair, using a half_gaussian subroutine

  end select

  end subroutine


  ! Sum new 1d shape into 1d taper
  subroutine  tapers_compile(add,taper_in,t1,t2)
    real   , intent(in)       :: taper_in(:)
    logical, intent(in)       :: add
    integer                   :: dir,t1,t2
    integer                                :: ii
    
    if (.not. size(taper_in) == n) then; write(0,*) "size(taper_in)!=size(taper) in next call to tapers_compile"; call exit(); end if      
    
    call addnull(add,taper_com(t1:t2))
    do ii=t1,t2;
      taper_com(ii) = taper_com(ii) + taper_in(ii)
    end do
    
  end subroutine


  ! Invert a shape  x: (max(x)-x)
  subroutine tapers_invert1 (inv,x)
    real,intent(inout) :: x(:)
    logical,intent(in) :: inv    
    
    if (inv) then
      x = max(maxval(x),1.)-x
    end if

  end subroutine

  ! Invert a shape  x: (max(x)-x)
  subroutine tapers_invert2 (inv,x)
    real,intent(inout) :: x(:,:)
    logical,intent(in) :: inv    
    
    if (inv) then
      x = max(maxval(x),1.)-x
    end if

  end subroutine

  ! Invert a shape  x: (max(x)-x)
  subroutine tapers_invert3 (inv,x)
    real,intent(inout) :: x(:,:,:)
    logical,intent(in) :: inv    
    
    if (inv) then
      x = max(maxval(x),1.)-x
    end if

  end subroutine

  ! Get an indice given an absolute time, d and o
  subroutine taper_indice(t,t_ind,d_in,o_in)
    real,optional,intent(in) :: d_in,o_in
    real                     :: d   ,o
    integer                  :: t_ind
    real                     :: t
    if (present(d_in)) then; d = d_in; else; d = 1.; end if
    if (present(o_in)) then; o = o_in; else; o = 0.; end if

    t_ind = int((t-o)/d+.5)+1

  end subroutine




  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  !     1D TAPER SHAPES
  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  
  subroutine tapers_zero(t1_in,t2_in)
    integer, optional, intent(in)          :: t1_in,t2_in
    integer                                :: t1,t2
    integer                                :: ii
    if (present(t1_in)) then; t1 = t1_in; else; t1=1; end if
    if (present(t2_in)) then; t2 = t2_in; else; t2=n; end if
    ! To zero the taper in the memory of mod_taper
    ! Without deallocating it

    call addnull       (.false.,taper)
    call tapers_compile(.false.,taper,t1,t2)
    
  end subroutine
  
  
  
  subroutine tapers_rect           (add,t1,t2,scale_in)
    logical, intent(in)          :: add
    integer, intent(in)          :: t1,t2
    real, optional,intent(in)    :: scale_in
    real                         :: scale
    integer                                :: ii

    call addnull(.false.,taper)
    if (present(scale_in)) then; scale = scale_in; else; scale=1.; end if

    ! Construction of taper
    taper(t1:t2) = 1.
    
    call tapers_compile( add,scale*taper,t1,t2 )
    
  end subroutine
  

  subroutine tapers_linear_up(add,t1,t2,scale_in)
    logical, intent(in)          :: add
    integer, intent(in)          :: t1,t2
    real, optional,intent(in)    :: scale_in
    real                         :: scale
      integer                                :: ii
  
    call addnull(.false.,taper)
    if (present(scale_in)) then; scale = scale_in; else; scale=1.; end if
    
    ! Construction of taper
    do ii=max(1,t1),min(t2,n)
      taper(ii) = real(ii-t1)/real(t2-t1)
    end do
    
    call tapers_compile( add,scale*taper,max(1,t1),min(t2,n) )
    
  end subroutine

  subroutine tapers_linear_down(add,t1,t2,scale_in)
    logical, intent(in)          :: add
    integer, intent(in)          :: t1,t2
    real, optional,intent(in)    :: scale_in
    real                         :: scale
    integer                                :: ii
    
    call addnull(.false.,taper)
    if (present(scale_in)) then; scale = scale_in; else; scale=1.; end if
    
    ! Construction of taper
    do ii=max(1,t1),min(t2,n)
      taper(ii) = 1.-real(ii-t1)/real(t2-t1)
    end do
    
    call tapers_compile( add,scale*taper,max(1,t1),min(t2,n) )
    
  end subroutine
  
  
  subroutine tapers_pyramid(add,t1,t2,scale_in)
    ! Combine two linear functions !
    logical, intent(in)          :: add
    integer, intent(in)          :: t1,t2
    real, optional,intent(in)    :: scale_in
    real                         :: scale
    integer                                :: ii
    
    call addnull(.false.,taper)
    if (present(scale_in)) then; scale = scale_in; else; scale=1.; end if
    
    ! Construction of taper
    call tapers_linear_up  (.false.,t1,t1 + (t2-t1)/2,scale_in=scale)
    call tapers_linear_down(.false.,t1 + (t2-t1)/2,t2,scale_in=scale)
    
  end subroutine
  
  subroutine tapers_pyramid_block(add,t1,t2,t3,t4,scale_in)
    ! Combine two linear functions and a rectangle !
    logical, intent(in)          :: add
    integer, intent(in)          :: t1,t2,t3,t4
    real, optional,intent(in)    :: scale_in
    real                         :: scale
    integer                                :: ii
    
    call addnull(.false.,taper)
    if (present(scale_in)) then; scale = scale_in; else; scale=1.; end if
    
    !write(0,*) n,t1,t2,t3,t4
    
    ! Construction of taper
    call tapers_linear_up  ( .false.,t1, t2,scale_in=scale )
    call tapers_rect       ( .false.,t2, t3,scale_in=scale )
    call tapers_linear_down( .false.,t3, t4,scale_in=scale )
    
  end subroutine
  
  
  subroutine tapers_hann           (add,t1,t2,scale_in)
    logical, intent(in)          :: add
    integer, intent(in)          :: t1,t2
    real, optional,intent(in)    :: scale_in
    real                         :: scale
    integer                                :: ii

    call addnull(.false.,taper)
    if (present(scale_in)) then; scale = scale_in; else; scale=1.; end if

    ! Construction of taper
    do ii=max(1,t1),min(t2,n)
      taper(ii) = .5*(1-cos(2.*pi*real(ii-t1)/real(t2-t1)))
    end do

    call tapers_compile( add,scale*taper,max(1,t1),min(t2,n) )
    
  end subroutine
  
  
  subroutine tapers_hann_half      (add,t1,t2,scale_in)
    ! Combine hann and rectangular (in that order arranged in time) !
    logical, intent(in)          :: add
    integer, intent(in)          :: t1,t2
    real, optional,intent(in)    :: scale_in
    real                         :: scale
    integer                      :: ii
    
    call addnull(.false.,taper)
    if (present(scale_in)) then; scale = scale_in; else; scale=1.; end if

    ! Construction of taper
    do ii=max(1,t1),min(t2,n)
      taper(ii) = .5*(1-cos(pi*real(ii-t1)/real(t2-t1)))
    end do
    call tapers_compile( add,scale*taper,max(1,t1),min(t2,n) )

    ! Suplement with a rectangle
    call tapers_rect(.false.,t2,            n ,scale_in=scale )

  end subroutine
  
  
  subroutine tapers_hann_block (add,t1,t2,t3,t4,scale_in)
    ! Combine two hann functions and a rectangle !
    logical, intent(in)          :: add
    integer, intent(in)          :: t1,t2,t3,t4
    real,optional,intent(in)     :: scale_in
    real                         :: scale
    integer                      :: ii

    call addnull(.false.,taper)
    if (present(scale_in)) then; scale = scale_in; else; scale=1.; end if
    
    ! Construction of taper
    do ii=max(1,t1),min(t2,n)
      taper(ii) = .5*( 1-cos(pi*real(ii-t1)/real(t2-t1)) )
    end do
    do ii=max(1,t3),min(t4,n)
      taper(ii) = .5*( 1-cos(pi*real(ii-t3-(t4-t3))/real(t4-t3)) )
    end do
    call tapers_compile( add,scale*taper,max(1,t1),min(t2,n) )
    call tapers_compile( add,scale*taper,max(1,t3),min(t4,n) )
    call tapers_rect   ( add, t2, t3             , scale_in=scale )
    
  end subroutine
  
  
  subroutine tapers_cosine        (add,t1,t2,scale_in)
    logical, intent(in)          :: add
    integer, intent(in)          :: t1,t2
    real, optional,intent(in)    :: scale_in
    real                         :: scale
    integer                      :: ii
    
    call addnull(.false.,taper)
    if (present(scale_in)) then; scale = scale_in; else; scale=1.; end if
    
    ! Construction of taper
    do ii=max(1,t1),min(t2,n)
      taper(ii) = sin(pi*(ii-t1)/(t2-t1+1))
    end do
    
    call tapers_compile( add,scale*taper,max(1,t1),min(t2,n) )
    
  end subroutine
  
  
  subroutine tapers_gaussian(add,t1,t2,scale_in,sigma_in)
    ! Spread argument makes the Gaussian spread beyond the t1-t2 interval,
    ! An appropriate sigma can also handle that.
    ! sigma_in < .5, and is set as such.
    logical, intent(in)          :: add
    integer, intent(in)          :: t1,t2
    real, optional,intent(in)    :: scale_in,sigma_in
    real                         :: scale,sigma
     integer                     :: ii
   
    call addnull(.false.,taper)
    if (present(scale_in)) then; scale = scale_in; else; scale = 1.; end if
    if (present(sigma_in)) then; sigma = sigma_in; else; sigma = .4; end if
    if ( sigma > .5 )      then; sigma = .5      ; end if

    ! Construction of taper
    do ii=max(1,t1),min(t2,n)
      taper(ii) = exp(-.5*((((ii-t1)-((t2-t1)/2))/(sigma*((t2-t1)/2)))**2))
    end do
    
    call tapers_compile( add,scale*taper,max(1,t1),min(t2,n) )
    
  end subroutine
  

  subroutine tapers_gaussian_half(add,t1,t2,scale_in,sigma_in)
    ! Spread argument makes the Gaussian spread beyond the t1-t2 interval,
    ! An appropriate sigma can also handle that.
    ! sigma_in < .5, and is set as such.

    ! Combine gaussian and rectangular (in that order arranged in time) !
    logical, intent(in)          :: add
    integer, intent(in)          :: t1,t2
    real, optional,intent(in)    :: scale_in,sigma_in
    real                         :: scale,sigma
    integer                      :: ii

    call addnull(.false.,taper)
    if (present(scale_in)) then; scale = scale_in; else; scale = 1.; end if
    if (present(sigma_in)) then; sigma = sigma_in; else; sigma = .4; end if
    if ( sigma > .5 )      then; sigma = .5      ; end if

    ! Construction of taper
    ! Construction of taper
    do ii=max(1,t1),min(t2,n)
      taper(ii) = exp(-.5*((((ii-t2))/(sigma*(t2-t1)))**2))
    end do
    call tapers_compile( add,scale*taper,max(1,t1),min(t2,n) )

    !call tapers_gaussian(.false.,t1,t1 + 2*(t2-t1),scale_in=scale,sigma_in=sigma)
    call tapers_rect    (.false.,t2,            n ,scale_in=scale)

  end subroutine


  subroutine tapers_gaussian_block(add,t1,t2,t3,t4,scale_in,sigma_in)
    ! Combine two hann functions and a rectangle !
    logical, intent(in)          :: add
    integer, intent(in)          :: t1,t2,t3,t4
    real, optional,intent(in)    :: scale_in,sigma_in
    real                         :: scale,sigma
    integer                      :: ii
    
    call addnull(.false.,taper)
    if (present(scale_in)) then; scale = scale_in; else; scale = 1.; end if
    if (present(sigma_in)) then; sigma = sigma_in; else; sigma = .4; end if
    if ( sigma > .5 )      then; sigma = .5      ; end if
    
    ! Construction of taper
    do ii=max(1,t1),min(t2,n)
      taper(ii) = exp(-.5*((((ii-t2))/(sigma*(t2-t1)))**2))
    end do

    do ii=max(1,t3),min(t4,n)
      taper(ii) = exp(-.5*((((ii-t3))/(sigma*(t4-t3)))**2))
    end do
    call tapers_compile( add,scale*taper,max(1,t1),min(t2,n) )
    call tapers_compile( add,scale*taper,max(1,t3),min(t4,n) )

    call tapers_rect    (.false., t2, t3,scale_in=scale )    
  end subroutine
  



end module
